import torch.nn as nn
import torch.nn.functional as F


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class BasicBlock(nn.Module):
    def __init__(self, in_chan, out_chan, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(in_chan, out_chan, stride)
        self.bn1 = nn.BatchNorm2d(out_chan)
        self.conv2 = conv3x3(out_chan, out_chan)
        self.bn2 = nn.BatchNorm2d(out_chan)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = None
        if in_chan != out_chan or stride != 1:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_chan, out_chan, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(out_chan),
            )

    def forward(self, x):
        residual = self.conv1(x)
        residual = F.relu(self.bn1(residual))
        residual = self.conv2(residual)
        residual = self.bn2(residual)

        shortcut = x
        if self.downsample is not None:
            shortcut = self.downsample(x)

        out = shortcut + residual
        out = self.relu(out)
        return out


def create_layer_basic(in_chan, out_chan, bnum, stride=1):
    layers = [BasicBlock(in_chan, out_chan, stride=stride)]
    for i in range(bnum - 1):
        layers.append(BasicBlock(out_chan, out_chan, stride=1))
    return nn.Sequential(*layers)


class ResNet18(nn.Module):
    def __init__(self):
        super(ResNet18, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = create_layer_basic(64, 64, bnum=2, stride=1)
        self.layer2 = create_layer_basic(64, 128, bnum=2, stride=2)
        self.layer3 = create_layer_basic(128, 256, bnum=2, stride=2)
        self.layer4 = create_layer_basic(256, 512, bnum=2, stride=2)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(self.bn1(x))
        x = self.maxpool(x)

        x = self.layer1(x)
        feat8 = self.layer2(x)  # 1/8
        feat16 = self.layer3(feat8)  # 1/16
        feat32 = self.layer4(feat16)  # 1/32
        return feat8, feat16, feat32



import uuid

from PIL import Image

from nataili.model_manager.compvis import CompVisModelManager
from nataili.model_manager.safety_checker import SafetyCheckerModelManager
from nataili.stable_diffusion.compvis import CompVis
from nataili.util.logger import logger

init_image = Image.open("./01.png").convert("RGB")

mm = CompVisModelManager()
smm = SafetyCheckerModelManager()

model = "stable_diffusion"
safety_model = "safety_checker"

def test_compvis(
    model,
    prompt,
    sampler,
    steps=30,
    output_dir=None,
    seed=None,
    init_img=None,
    denoising_strength=0.75,
    sigma_override=None,
    filter_nsfw=False,
    safety_checker=None,
):
    log_message = f"sampler: {sampler} steps: {steps} model: {model}"
    if seed:
        log_message += f" seed: {seed}"
    if sigma_override:
        log_message += f" sigma_override: {sigma_override}"
    logger.info(log_message)
    if filter_nsfw:
        logger.info("Filtering NSFW")
    if init_img:
        logger.info("Using init image")
        logger.info(f"Denoising strength: {denoising_strength}")
    compvis = CompVis(
        model=mm.loaded_models[model],
        model_name=model,
        output_dir=output_dir,
        disable_voodoo=True,
        filter_nsfw=filter_nsfw,
        safety_checker=safety_checker,
    )
    compvis.generate(
        prompt,
        sampler_name=sampler,
        ddim_steps=steps,
        seed=seed,
        init_img=init_img,
        sigma_override=sigma_override,
        denoising_strength=denoising_strength,
    )


samplers = [
    "k_lms",
]

output_dir = f"./test_output/{str(uuid.uuid4())}"


@logger.catch
def test():
    mm.load(model)
    smm.load(safety_model)
    logger.info(f"Output dir: {output_dir}")
    logger.debug(f"Running inference on {model}")
    logger.info(f"Testing safety checker")
    prompt = (
        "boobs"
    )
    logger.info(f"Prompt: {prompt}")
    for sampler in samplers:
        test_compvis(model, prompt, sampler, output_dir=output_dir, safety_checker=smm.loaded_models[safety_model]["model"], filter_nsfw=True)

if __name__ == "__main__":
    test()



import math
import numpy as np
import matplotlib
import cv2


def padRightDownCorner(img, stride, padValue):
    h = img.shape[0]
    w = img.shape[1]

    pad = 4 * [None]
    pad[0] = 0 # up
    pad[1] = 0 # left
    pad[2] = 0 if (h % stride == 0) else stride - (h % stride) # down
    pad[3] = 0 if (w % stride == 0) else stride - (w % stride) # right

    img_padded = img
    pad_up = np.tile(img_padded[0:1, :, :]*0 + padValue, (pad[0], 1, 1))
    img_padded = np.concatenate((pad_up, img_padded), axis=0)
    pad_left = np.tile(img_padded[:, 0:1, :]*0 + padValue, (1, pad[1], 1))
    img_padded = np.concatenate((pad_left, img_padded), axis=1)
    pad_down = np.tile(img_padded[-2:-1, :, :]*0 + padValue, (pad[2], 1, 1))
    img_padded = np.concatenate((img_padded, pad_down), axis=0)
    pad_right = np.tile(img_padded[:, -2:-1, :]*0 + padValue, (1, pad[3], 1))
    img_padded = np.concatenate((img_padded, pad_right), axis=1)

    return img_padded, pad

# transfer caffe model to pytorch which will match the layer name
def transfer(model, model_weights):
    transfered_model_weights = {}
    for weights_name in model.state_dict().keys():
        transfered_model_weights[weights_name] = model_weights['.'.join(weights_name.split('.')[1:])]
    return transfered_model_weights

# draw the body keypoint and lims
def draw_bodypose(canvas, candidate, subset):
    stickwidth = 4
    limbSeq = [[2, 3], [2, 6], [3, 4], [4, 5], [6, 7], [7, 8], [2, 9], [9, 10], \
               [10, 11], [2, 12], [12, 13], [13, 14], [2, 1], [1, 15], [15, 17], \
               [1, 16], [16, 18], [3, 17], [6, 18]]

    colors = [[255, 0, 0], [255, 85, 0], [255, 170, 0], [255, 255, 0], [170, 255, 0], [85, 255, 0], [0, 255, 0], \
              [0, 255, 85], [0, 255, 170], [0, 255, 255], [0, 170, 255], [0, 85, 255], [0, 0, 255], [85, 0, 255], \
              [170, 0, 255], [255, 0, 255], [255, 0, 170], [255, 0, 85]]
    for i in range(18):
        for n in range(len(subset)):
            index = int(subset[n][i])
            if index == -1:
                continue
            x, y = candidate[index][0:2]
            cv2.circle(canvas, (int(x), int(y)), 4, colors[i], thickness=-1)
    for i in range(17):
        for n in range(len(subset)):
            index = subset[n][np.array(limbSeq[i]) - 1]
            if -1 in index:
                continue
            cur_canvas = canvas.copy()
            Y = candidate[index.astype(int), 0]
            X = candidate[index.astype(int), 1]
            mX = np.mean(X)
            mY = np.mean(Y)
            length = ((X[0] - X[1]) ** 2 + (Y[0] - Y[1]) ** 2) ** 0.5
            angle = math.degrees(math.atan2(X[0] - X[1], Y[0] - Y[1]))
            polygon = cv2.ellipse2Poly((int(mY), int(mX)), (int(length / 2), stickwidth), int(angle), 0, 360, 1)
            cv2.fillConvexPoly(cur_canvas, polygon, colors[i])
            canvas = cv2.addWeighted(canvas, 0.4, cur_canvas, 0.6, 0)
    # plt.imsave("preview.jpg", canvas[:, :, [2, 1, 0]])
    # plt.imshow(canvas[:, :, [2, 1, 0]])
    return canvas


# image drawed by opencv is not good.
def draw_handpose(canvas, all_hand_peaks, show_number=False):
    edges = [[0, 1], [1, 2], [2, 3], [3, 4], [0, 5], [5, 6], [6, 7], [7, 8], [0, 9], [9, 10], \
             [10, 11], [11, 12], [0, 13], [13, 14], [14, 15], [15, 16], [0, 17], [17, 18], [18, 19], [19, 20]]

    for peaks in all_hand_peaks:
        for ie, e in enumerate(edges):
            if np.sum(np.all(peaks[e], axis=1)==0)==0:
                x1, y1 = peaks[e[0]]
                x2, y2 = peaks[e[1]]
                cv2.line(canvas, (x1, y1), (x2, y2), matplotlib.colors.hsv_to_rgb([ie/float(len(edges)), 1.0, 1.0])*255, thickness=2)

        for i, keyponit in enumerate(peaks):
            x, y = keyponit
            cv2.circle(canvas, (x, y), 4, (0, 0, 255), thickness=-1)
            if show_number:
                cv2.putText(canvas, str(i), (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.3, (0, 0, 0), lineType=cv2.LINE_AA)
    return canvas

# detect hand according to body pose keypoints
# please refer to https://github.com/CMU-Perceptual-Computing-Lab/openpose/blob/master/src/openpose/hand/handDetector.cpp
def handDetect(candidate, subset, oriImg):
    # right hand: wrist 4, elbow 3, shoulder 2
    # left hand: wrist 7, elbow 6, shoulder 5
    ratioWristElbow = 0.33
    detect_result = []
    image_height, image_width = oriImg.shape[0:2]
    for person in subset.astype(int):
        # if any of three not detected
        has_left = np.sum(person[[5, 6, 7]] == -1) == 0
        has_right = np.sum(person[[2, 3, 4]] == -1) == 0
        if not (has_left or has_right):
            continue
        hands = []
        #left hand
        if has_left:
            left_shoulder_index, left_elbow_index, left_wrist_index = person[[5, 6, 7]]
            x1, y1 = candidate[left_shoulder_index][:2]
            x2, y2 = candidate[left_elbow_index][:2]
            x3, y3 = candidate[left_wrist_index][:2]
            hands.append([x1, y1, x2, y2, x3, y3, True])
        # right hand
        if has_right:
            right_shoulder_index, right_elbow_index, right_wrist_index = person[[2, 3, 4]]
            x1, y1 = candidate[right_shoulder_index][:2]
            x2, y2 = candidate[right_elbow_index][:2]
            x3, y3 = candidate[right_wrist_index][:2]
            hands.append([x1, y1, x2, y2, x3, y3, False])

        for x1, y1, x2, y2, x3, y3, is_left in hands:
            # pos_hand = pos_wrist + ratio * (pos_wrist - pos_elbox) = (1 + ratio) * pos_wrist - ratio * pos_elbox
            # handRectangle.x = posePtr[wrist*3] + ratioWristElbow * (posePtr[wrist*3] - posePtr[elbow*3]);
            # handRectangle.y = posePtr[wrist*3+1] + ratioWristElbow * (posePtr[wrist*3+1] - posePtr[elbow*3+1]);
            # const auto distanceWristElbow = getDistance(poseKeypoints, person, wrist, elbow);
            # const auto distanceElbowShoulder = getDistance(poseKeypoints, person, elbow, shoulder);
            # handRectangle.width = 1.5f * fastMax(distanceWristElbow, 0.9f * distanceElbowShoulder);
            x = x3 + ratioWristElbow * (x3 - x2)
            y = y3 + ratioWristElbow * (y3 - y2)
            distanceWristElbow = math.sqrt((x3 - x2) ** 2 + (y3 - y2) ** 2)
            distanceElbowShoulder = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
            width = 1.5 * max(distanceWristElbow, 0.9 * distanceElbowShoulder)
            # x-y refers to the center --> offset to topLeft point
            # handRectangle.x -= handRectangle.width / 2.f;
            # handRectangle.y -= handRectangle.height / 2.f;
            x -= width / 2
            y -= width / 2  # width = height
            # overflow the image
            if x < 0: x = 0
            if y < 0: y = 0
            width1 = width
            width2 = width
            if x + width > image_width: width1 = image_width - x
            if y + width > image_height: width2 = image_height - y
            width = min(width1, width2)
            # the max hand box value is 20 pixels
            if width >= 20:
                detect_result.append([int(x), int(y), int(width), is_left])

    '''
    return value: [[x, y, w, True if left hand else False]].
    width=height since the network require squared input.
    x, y is the coordinate of top left 
    '''
    return detect_result

# get max index of 2d array
def npmax(array):
    arrayindex = array.argmax(1)
    arrayvalue = array.max(1)
    i = arrayvalue.argmax()
    j = arrayindex[i]
    return i, j



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
import time
from pathlib import Path

import open_clip
import torch
import transformers.utils.hub
from omegaconf import OmegaConf
from torch import nn

import ldm.modules.encoders.modules
from ldm.util import instantiate_from_config
from nataili import enable_ray_alternative
from nataili.cache import get_cache_directory
from nataili.model_manager.base import BaseModelManager
from nataili.util.logger import logger
from nataili.util.voodoo import get_model_cache_filename, have_model_cache, push_model_to_plasma


class CompVisModelManager(BaseModelManager):
    def __init__(self, download_reference=True, custom_path="models/custom"):
        super().__init__()
        self.download_reference = download_reference
        self.path = f"{get_cache_directory()}/compvis"
        self.custom_path = custom_path
        self.models_db_name = "stable_diffusion"
        self.models_path = self.pkg / f"{self.models_db_name}.json"
        self.remote_db = (
            f"https://raw.githubusercontent.com/db0/AI-Horde-image-model-reference/main/{self.models_db_name}.json"
        )
        self.init()

    def load(
        self,
        model_name: str,
        half_precision=True,
        gpu_id=0,
        cpu_only=False,
        voodoo=False,
    ):
        """
        model_name: str. Name of the model to load. See available_models for a list of available models.
        half_precision: bool. If True, the model will be loaded in half precision.
        gpu_id: int. The id of the gpu to use. If the gpu is not available, the model will be loaded on the cpu.
        cpu_only: bool. If True, the model will be loaded on the cpu. If True, half_precision will be set to False.
        voodoo: bool. Voodoo (Ray)
        """
        if model_name not in self.models:
            logger.error(f"{model_name} not found")
            return False
        if model_name not in self.available_models:
            logger.error(f"{model_name} not available")
            self.download_model(model_name)
            logger.init_ok(f"{model_name}", status="Downloaded")
        if model_name not in self.loaded_models:
            if not self.cuda_available:
                cpu_only = True
                voodoo = False
            tic = time.time()
            logger.init(f"{model_name}", status="Loading")
            self.loaded_models[model_name] = self.load_compvis(
                model_name,
                half_precision=half_precision,
                gpu_id=gpu_id,
                cpu_only=cpu_only,
                voodoo=voodoo,
            )
            toc = time.time()
            logger.init_ok(f"{model_name}: {round(toc-tic,2)} seconds", status="Loaded")
            return True

    def load_model_from_config(self, model_path="", config_path="", map_location="cpu"):
        config = OmegaConf.load(config_path)
        pl_sd = torch.load(model_path, map_location=map_location)
        if "global_step" in pl_sd:
            logger.info(f"Global Step: {pl_sd['global_step']}")
        sd = pl_sd["state_dict"] if "state_dict" in pl_sd else pl_sd

        sd1_clip_weight = "cond_stage_model.transformer.text_model.embeddings.token_embedding.weight"
        sd2_clip_weight = "cond_stage_model.model.transformer.resblocks.0.attn.in_proj_weight"
        clip_is_included_into_sd = sd1_clip_weight in sd or sd2_clip_weight in sd

        model = None
        try:
            with DisableInitialization(disable_clip=clip_is_included_into_sd):
                model = instantiate_from_config(config.model)
        except Exception as e:
            pass

        if model is None:
            logger.info("Failed to create model quickly; will retry using slow method.")
            model = instantiate_from_config(config.model)

        m, u = model.load_state_dict(sd, strict=False)
        model = model.eval()
        for m in model.modules():
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                m._orig_padding_mode = m.padding_mode
        del pl_sd, sd, m, u
        return model

    def load_custom(self, ckpt_path, config_path, model_name=None, replace=False):
        if not os.path.isfile(ckpt_path):
            logger.error(f"{ckpt_path} not found")
            return
        if not os.path.isfile(config_path):
            logger.error(f"{config_path} not found")
            return
        if not ckpt_path.endswith(".ckpt"):
            logger.error(f"{ckpt_path} is not a valid checkpoint file")
            return
        if not config_path.endswith(".yaml"):
            logger.error(f"{config_path} is not a valid config file")
            return
        if model_name is None:
            model_name = os.path.basename(ckpt_path).replace(".ckpt", "")
        if model_name not in self.models or replace:
            self.models[model_name] = {
                "name": model_name,
                "type": "ckpt",
                "description": f"custom model {model_name}",
                "config": {
                    "files": [
                        {"path": f"{self.custom_path}/{model_name}.ckpt"},
                        {"path": f"{self.custom_path}/{model_name}.yaml"},
                    ]
                },
                "available": True,
            }
            self.available_models.append(model_name)

    def load_available_models_from_custom(self, replace=False):
        # ckpt files and matching config yaml files
        for file in os.listdir(self.custom_path):
            if file.endswith(".ckpt"):
                ckpt_path = f"{self.custom_path}/{file}"
                config_path = ckpt_path.replace(".ckpt", ".yaml")
                self.load_custom(
                    ckpt_path,
                    config_path,
                    replace=replace,
                )

    def load_compvis(
        self,
        model_name,
        half_precision=True,
        gpu_id=0,
        cpu_only=False,
        voodoo=False,
    ):
        ckpt_path = self.get_model_files(model_name)[0]["path"]
        ckpt_path = f"{self.path}/{ckpt_path}"
        config_path = self.get_model_files(model_name)[1]["path"]
        config_path = f"{self.pkg}/{config_path}"
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        logger.debug(f"Loading model {model_name} on {device}")
        logger.debug(f"Model path: {ckpt_path}")

        if voodoo and enable_ray_alternative.active and have_model_cache(ckpt_path):
            logger.debug("Have up to date model cache, using that instead of ckpt model")
            model = get_model_cache_filename(ckpt_path)
        else:
            logger.debug("Loading model from checkpoint")
            model = self.load_model_from_config(model_path=ckpt_path, config_path=config_path)
            if half_precision:
                logger.debug("Converting model to half precision")
                model = model.half()
                logger.debug("Converting model.cond_stage_model.transformer to half precision")
                if "stable diffusion 2" in self.models[model_name]["baseline"]:
                    model.cond_stage_model.model.transformer = model.cond_stage_model.model.transformer.half()
                else:
                    model.cond_stage_model.transformer = model.cond_stage_model.transformer.half()

        if voodoo and isinstance(model, torch.nn.Module):
            logger.debug(f"Doing voodoo on {model_name}")
            model = push_model_to_plasma(model, ckpt_path)
        elif isinstance(model, torch.nn.Module):
            logger.debug(f"Moving model data directly to device {device}")
            model = model.to(device)
            logger.debug(f"Sending model.cond_stage_model.transformer to {device}")
            if "stable diffusion 2" in self.models[model_name]["baseline"]:
                model.cond_stage_model.model.transformer = model.cond_stage_model.model.transformer.to(device)
            else:
                model.cond_stage_model.transformer = model.cond_stage_model.transformer.to(device)
            logger.debug(f"Setting model.cond_stage_model.device to {device}")
            model.cond_stage_model.device = device

        return {"model": model, "device": device, "half_precision": half_precision}

    def check_model_available(self, model_name):
        if model_name not in self.models:
            return False
        return self.check_file_available(self.get_model_files(model_name)[0]["path"])


class DisableInitialization:
    """
    When an object of this class enters a `with` block, it starts:
    - preventing torch's layer initialization functions from working
    - changes CLIP and OpenCLIP to not download model weights
    - changes CLIP to not make requests to check if there is a new version of a file you already have
    When it leaves the block, it reverts everything to how it was before.
    Use it like this:
    ```
    with DisableInitialization():
        do_things()
    ```
    """

    def __init__(self, disable_clip=True):
        self.replaced = []
        self.disable_clip = disable_clip

    def replace(self, obj, field, func):
        original = getattr(obj, field, None)
        if original is None:
            return None

        self.replaced.append((obj, field, original))
        setattr(obj, field, func)

        return original

    def __enter__(self):
        def do_nothing(*args, **kwargs):
            pass

        def create_model_and_transforms_without_pretrained(*args, pretrained=None, **kwargs):
            return self.create_model_and_transforms(*args, pretrained=None, **kwargs)

        def CLIPTextModel_from_pretrained(pretrained_model_name_or_path, *model_args, **kwargs):
            if sys.version_info != (3, 8, 10):
                try:
                    res = self.CLIPTextModel_from_pretrained(
                        None, *model_args, config=pretrained_model_name_or_path, state_dict={}, **kwargs
                    )
                except Exception as e:
                    res = self.CLIPTextModel_from_pretrained(None, *model_args, **kwargs)
            else:
                res = self.CLIPTextModel_from_pretrained(None, *model_args, **kwargs)

            res.name_or_path = pretrained_model_name_or_path
            return res

        def transformers_modeling_utils_load_pretrained_model(*args, **kwargs):
            args = (
                args[0:3] + ("/",) + args[4:]
            )  # resolved_archive_file; must set it to something to prevent what seems to be a bug
            return self.transformers_modeling_utils_load_pretrained_model(*args, **kwargs)

        def transformers_utils_hub_get_file_from_cache(original, url, *args, **kwargs):
            # this file is always 404, prevent making request
            if (
                url == "https://huggingface.co/openai/clip-vit-large-patch14/resolve/main/added_tokens.json"
                or url == "openai/clip-vit-large-patch14"
                and args[0] == "added_tokens.json"
            ):
                return None

            try:
                res = original(url, *args, local_files_only=True, **kwargs)
                if res is None:
                    res = original(url, *args, local_files_only=False, **kwargs)
                return res
            except Exception as e:
                return original(url, *args, local_files_only=False, **kwargs)

        def transformers_utils_hub_get_from_cache(url, *args, local_files_only=False, **kwargs):
            return transformers_utils_hub_get_file_from_cache(
                self.transformers_utils_hub_get_from_cache, url, *args, **kwargs
            )

        def transformers_tokenization_utils_base_cached_file(url, *args, local_files_only=False, **kwargs):
            return transformers_utils_hub_get_file_from_cache(
                self.transformers_tokenization_utils_base_cached_file, url, *args, **kwargs
            )

        def transformers_configuration_utils_cached_file(url, *args, local_files_only=False, **kwargs):
            return transformers_utils_hub_get_file_from_cache(
                self.transformers_configuration_utils_cached_file, url, *args, **kwargs
            )

        self.replace(torch.nn.init, "kaiming_uniform_", do_nothing)
        self.replace(torch.nn.init, "_no_grad_normal_", do_nothing)
        self.replace(torch.nn.init, "_no_grad_uniform_", do_nothing)

        if self.disable_clip:
            self.create_model_and_transforms = self.replace(
                open_clip, "create_model_and_transforms", create_model_and_transforms_without_pretrained
            )
            self.CLIPTextModel_from_pretrained = self.replace(
                ldm.modules.encoders.modules.CLIPTextModel, "from_pretrained", CLIPTextModel_from_pretrained
            )
            self.transformers_modeling_utils_load_pretrained_model = self.replace(
                transformers.modeling_utils.PreTrainedModel,
                "_load_pretrained_model",
                transformers_modeling_utils_load_pretrained_model,
            )
            self.transformers_tokenization_utils_base_cached_file = self.replace(
                transformers.tokenization_utils_base, "cached_file", transformers_tokenization_utils_base_cached_file
            )
            self.transformers_configuration_utils_cached_file = self.replace(
                transformers.configuration_utils, "cached_file", transformers_configuration_utils_cached_file
            )
            self.transformers_utils_hub_get_from_cache = self.replace(
                transformers.utils.hub, "get_from_cache", transformers_utils_hub_get_from_cache
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        for obj, field, original in self.replaced:
            setattr(obj, field, original)

        self.replaced.clear()



from .drop import DropPath
from .inverted_residual import InvertedResidual, InvertedResidualV3
from .make_divisible import make_divisible
from .res_layer import ResLayer
from .se_layer import SELayer
from .self_attention_block import SelfAttentionBlock
from .up_conv_block import UpConvBlock
from .weight_init import trunc_normal_

__all__ = [
    'ResLayer', 'SelfAttentionBlock', 'make_divisible', 'InvertedResidual',
    'UpConvBlock', 'InvertedResidualV3', 'SELayer', 'DropPath', 'trunc_normal_'
]



import os.path as osp

import annotator.uniformer.mmcv as mmcv
import numpy as np

from ..builder import PIPELINES


@PIPELINES.register_module()
class LoadImageFromFile(object):
    """Load an image from file.

    Required keys are "img_prefix" and "img_info" (a dict that must contain the
    key "filename"). Added or updated keys are "filename", "img", "img_shape",
    "ori_shape" (same as `img_shape`), "pad_shape" (same as `img_shape`),
    "scale_factor" (1.0) and "img_norm_cfg" (means=0 and stds=1).

    Args:
        to_float32 (bool): Whether to convert the loaded image to a float32
            numpy array. If set to False, the loaded image is an uint8 array.
            Defaults to False.
        color_type (str): The flag argument for :func:`mmcv.imfrombytes`.
            Defaults to 'color'.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmcv.fileio.FileClient` for details.
            Defaults to ``dict(backend='disk')``.
        imdecode_backend (str): Backend for :func:`mmcv.imdecode`. Default:
            'cv2'
    """

    def __init__(self,
                 to_float32=False,
                 color_type='color',
                 file_client_args=dict(backend='disk'),
                 imdecode_backend='cv2'):
        self.to_float32 = to_float32
        self.color_type = color_type
        self.file_client_args = file_client_args.copy()
        self.file_client = None
        self.imdecode_backend = imdecode_backend

    def __call__(self, results):
        """Call functions to load image and get image meta information.

        Args:
            results (dict): Result dict from :obj:`mmseg.CustomDataset`.

        Returns:
            dict: The dict contains loaded image and meta information.
        """

        if self.file_client is None:
            self.file_client = mmcv.FileClient(**self.file_client_args)

        if results.get('img_prefix') is not None:
            filename = osp.join(results['img_prefix'],
                                results['img_info']['filename'])
        else:
            filename = results['img_info']['filename']
        img_bytes = self.file_client.get(filename)
        img = mmcv.imfrombytes(
            img_bytes, flag=self.color_type, backend=self.imdecode_backend)
        if self.to_float32:
            img = img.astype(np.float32)

        results['filename'] = filename
        results['ori_filename'] = results['img_info']['filename']
        results['img'] = img
        results['img_shape'] = img.shape
        results['ori_shape'] = img.shape
        # Set initial values for default meta_keys
        results['pad_shape'] = img.shape
        results['scale_factor'] = 1.0
        num_channels = 1 if len(img.shape) < 3 else img.shape[2]
        results['img_norm_cfg'] = dict(
            mean=np.zeros(num_channels, dtype=np.float32),
            std=np.ones(num_channels, dtype=np.float32),
            to_rgb=False)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(to_float32={self.to_float32},'
        repr_str += f"color_type='{self.color_type}',"
        repr_str += f"imdecode_backend='{self.imdecode_backend}')"
        return repr_str


@PIPELINES.register_module()
class LoadAnnotations(object):
    """Load annotations for semantic segmentation.

    Args:
        reduce_zero_label (bool): Whether reduce all label value by 1.
            Usually used for datasets where 0 is background label.
            Default: False.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmcv.fileio.FileClient` for details.
            Defaults to ``dict(backend='disk')``.
        imdecode_backend (str): Backend for :func:`mmcv.imdecode`. Default:
            'pillow'
    """

    def __init__(self,
                 reduce_zero_label=False,
                 file_client_args=dict(backend='disk'),
                 imdecode_backend='pillow'):
        self.reduce_zero_label = reduce_zero_label
        self.file_client_args = file_client_args.copy()
        self.file_client = None
        self.imdecode_backend = imdecode_backend

    def __call__(self, results):
        """Call function to load multiple types annotations.

        Args:
            results (dict): Result dict from :obj:`mmseg.CustomDataset`.

        Returns:
            dict: The dict contains loaded semantic segmentation annotations.
        """

        if self.file_client is None:
            self.file_client = mmcv.FileClient(**self.file_client_args)

        if results.get('seg_prefix', None) is not None:
            filename = osp.join(results['seg_prefix'],
                                results['ann_info']['seg_map'])
        else:
            filename = results['ann_info']['seg_map']
        img_bytes = self.file_client.get(filename)
        gt_semantic_seg = mmcv.imfrombytes(
            img_bytes, flag='unchanged',
            backend=self.imdecode_backend).squeeze().astype(np.uint8)
        # modify if custom classes
        if results.get('label_map', None) is not None:
            for old_id, new_id in results['label_map'].items():
                gt_semantic_seg[gt_semantic_seg == old_id] = new_id
        # reduce zero_label
        if self.reduce_zero_label:
            # avoid using underflow conversion
            gt_semantic_seg[gt_semantic_seg == 0] = 255
            gt_semantic_seg = gt_semantic_seg - 1
            gt_semantic_seg[gt_semantic_seg == 254] = 255
        results['gt_semantic_seg'] = gt_semantic_seg
        results['seg_fields'].append('gt_semantic_seg')
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(reduce_zero_label={self.reduce_zero_label},'
        repr_str += f"imdecode_backend='{self.imdecode_backend}')"
        return repr_str



import uuid

from PIL import Image

from nataili.model_manager.compvis import CompVisModelManager
from nataili.stable_diffusion.compvis import CompVis
from nataili.util.logger import logger

mm = CompVisModelManager()

model = "stable_diffusion"

mm.load(model, voodoo=True)


output_dir = f"./test_output/{str(uuid.uuid4())}"

prompt = "(stylized art:0.75), (Dungeons and Dragons:0.9), male (half-orc:1.1) barbarian, (wearing shorts and leather boots:1.05), (bald:1.1), (gray skin color:1.1), (holding a giant axe:1.2), (in the forest:1.1), (Kratos vibes:0.85), (Brock Lesnar vibes:0.95), (Drax vibes:0.9), wild beastly demeanor, (badass brute:1.05), art by Vox Machina and Code Lyoko and Maximum Ride, cel shaded, cinematic, 16k ### (low quality:0.85), bad proportions, (ornate:0.85), collage, multiple photos, text, watermark, group photo, crowded, mosh pit, yellow skin, (achromatic:0.75) ### weird mouth,weird teeth,incorrect anatomy, bad anatomy, ugly, odd, hideous, unsightly, extra limbs, poorly drawn, poorly drawn face, bad drawing, sketch, disfigured, cropped, out of frame,fully clothed,sad,agony,pain,cut off,trimmed,headpiece, head necklace,out of frame, malformed,hideous face, too many limbs, missing fingers, too many fingers, text, bad drawing, sketch, incorrect anatomy, hideous, unsightly, extra limbs, poorly drawn, poorly drawn face, disfigured, cropped"

compvis = CompVis(
    model=mm.loaded_models[model],
    model_name=model,
    output_dir=output_dir,
    disable_voodoo=False,
)
compvis.generate(
    prompt,
    hires_fix=True,
    width=768,
    height=576,
)



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import os

# threading
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from uuid import uuid4

import numpy as np
import torch
from PIL import Image

from nataili.cache import Cache
from nataili.util.cast import autocast_cuda
from nataili.util.logger import logger


class ImageEmbed:
    def __init__(self, model, cache: Cache):
        """
        :param model: Loaded model from ModelManager
        :param cache: Cache object
        """
        self.model = model
        self.cache = cache
        self.executor = ThreadPoolExecutor(max_workers=1024, thread_name_prefix="SaveThread")

    @autocast_cuda
    def _batch(self, pil_images: list):
        # logger.info(pil_images)
        for pil_image in pil_images:
            # logger.info(pil_image)
            pil_image["hash"] = hashlib.sha256(pil_image["pil_image"].tobytes()).hexdigest()
        preprocess_images = []
        to_remove = []
        with torch.no_grad():
            for pil_image in pil_images:
                try:
                    preprocess_images.append(
                        self.model["preprocess"](pil_image["pil_image"]).unsqueeze(0).to(self.model["device"])
                    )
                except RuntimeError as e:
                    logger.error(e)
                    logger.error(pil_image)
                    to_remove.append(pil_image)
                    continue
            for pil_image in to_remove:
                pil_images.remove(pil_image)
            assert len(preprocess_images) == len(pil_images)
            if len(preprocess_images) == 0:
                return
            preprocess_images = torch.cat(preprocess_images, dim=0)
            image_features = self.model["model"].encode_image(preprocess_images)
            for image_embed_array, pil_image in zip(image_features, pil_images):
                future = self.executor.submit(self._save, image_embed_array, pil_image["hash"])
                self.cache.add_sqlite_row(
                    file=pil_image["filename"].replace(".webp", ""), pil_hash=pil_image["hash"], hash=None
                )

    def _save(self, image_embed_array, image_hash):
        image_embed_array /= image_embed_array.norm(dim=-1, keepdim=True)
        np.save(f"{self.cache.cache_dir}/{image_hash}", image_embed_array.float().cpu().detach().numpy())

    @autocast_cuda
    def __call__(
        self, image: Image.Image = None, filename: str = None, directory: str = None, skip_cache: bool = False
    ):
        """
        :param pil_image: PIL image to embed
        SHA256 hash of image is used as key in cache
        If image is not in cache, embed it and save it to cache
        Returns SHA256 hash of image
        """
        if image is None and filename is None:
            raise ValueError("Either image or filename must be set")
        if image is not None and filename is not None:
            raise ValueError("Only one of image or filename must be set")
        if image is None:
            pil_image = Image.open(f"{directory}/{filename}").convert("RGB")
        else:
            pil_image = image
        if image is None:
            file_hash = hashlib.sha256(open(f"{directory}/{filename}", "rb").read()).hexdigest()
            image_hash = hashlib.sha256(pil_image.tobytes()).hexdigest()
        else:
            file_hash = None
            image_hash = hashlib.sha256(pil_image.tobytes()).hexdigest()
        if not skip_cache:
            cached = self.cache.get(pil_hash=image_hash)
            if cached:
                logger.debug(f"Image {image_hash} already in cache")
                return image_hash
        else:
            logger.debug(f"Skipping cache for image {image_hash}")
        logger.debug(f"Embedding image {image_hash}")
        with torch.no_grad():
            preprocess_image = self.model["preprocess"](pil_image).unsqueeze(0).to(self.model["device"])
        image_features = self.model["model"].encode_image(preprocess_image).float()
        self._save(image_features, image_hash)
        self.cache.add_sqlite_row(file=filename, hash=file_hash, pil_hash=image_hash)
        return image_hash



import sys
from functools import partialmethod

from loguru import logger

STDOUT_LEVELS = ["GENERATION", "PROMPT"]
INIT_LEVELS = ["INIT", "INIT_OK", "INIT_WARN", "INIT_ERR"]
MESSAGE_LEVELS = ["MESSAGE"]
STATS_LEVELS = ["STATS"]
# By default we're at error level or higher
verbosity = 20
quiet = 0


def set_logger_verbosity(count):
    global verbosity
    # The count comes reversed. So count = 0 means minimum verbosity
    # While count 5 means maximum verbosity
    # So the more count we have, the lowe we drop the versbosity maximum
    verbosity = 20 - (count * 10)


def quiesce_logger(count):
    global quiet
    # The bigger the count, the more silent we want our logger
    quiet = count * 10


def is_stdout_log(record):
    if record["level"].name not in STDOUT_LEVELS:
        return False
    if record["level"].no < verbosity + quiet:
        return False
    return True


def is_init_log(record):
    if record["level"].name not in INIT_LEVELS:
        return False
    if record["level"].no < verbosity + quiet:
        return False
    return True


def is_msg_log(record):
    if record["level"].name not in MESSAGE_LEVELS:
        return False
    if record["level"].no < verbosity + quiet:
        return False
    return True


def is_stderr_log(record):
    if record["level"].name in STDOUT_LEVELS + INIT_LEVELS + MESSAGE_LEVELS + STATS_LEVELS:
        return False
    if record["level"].no < verbosity + quiet:
        return False
    return True


def is_stats_log(record):
    if record["level"].name not in STATS_LEVELS:
        return False
    return True


def is_not_stats_log(record):
    if record["level"].name in STATS_LEVELS:
        return False
    return True


def is_trace_log(record):
    if record["level"].name not in ["TRACE", "ERROR"]:
        return False
    return True


def test_logger():
    logger.generation(
        "This is a generation message\nIt is typically multiline\nThee Lines".encode("unicode_escape").decode("utf-8")
    )
    logger.prompt("This is a prompt message")
    logger.debug("Debug Message")
    logger.info("Info Message")
    logger.warning("Info Warning")
    logger.error("Error Message")
    logger.critical("Critical Message")
    logger.init("This is an init message", status="Starting")
    logger.init_ok("This is an init message", status="OK")
    logger.init_warn("This is an init message", status="Warning")
    logger.init_err("This is an init message", status="Error")
    logger.message("This is user message")
    sys.exit()


logfmt = (
    "<level>{level: <10}</level> | <green>{time:YYYY-MM-DD HH:mm:ss.SSSSSS}</green> | "
    "<green>{name}</green>:<green>{function}</green>:<green>{line}</green> - <level>{message}</level>"
)
genfmt = "<level>{level: <10}</level> @ <green>{time:YYYY-MM-DD HH:mm:ss.SSSSSS}</green> | <level>{message}</level>"
initfmt = "<magenta>INIT      </magenta> | <level>{extra[status]: <11}</level> | <magenta>{message}</magenta>"
msgfmt = "<level>{level: <10}</level> | <level>{message}</level>"

try:
    logger.level("GENERATION", no=24, color="<cyan>")
    logger.level("PROMPT", no=23, color="<yellow>")
    logger.level("INIT", no=31, color="<white>")
    logger.level("INIT_OK", no=31, color="<green>")
    logger.level("INIT_WARN", no=31, color="<yellow>")
    logger.level("INIT_ERR", no=31, color="<red>")
    # Messages contain important information without which this application might not be able to be used
    # As such, they have the highest priority
    logger.level("MESSAGE", no=61, color="<green>")
    # Stats are info that might not display well in the terminal
    logger.level("STATS", no=19, color="<blue>")
except TypeError:
    pass

logger.__class__.generation = partialmethod(logger.__class__.log, "GENERATION")
logger.__class__.prompt = partialmethod(logger.__class__.log, "PROMPT")
logger.__class__.init = partialmethod(logger.__class__.log, "INIT")
logger.__class__.init_ok = partialmethod(logger.__class__.log, "INIT_OK")
logger.__class__.init_warn = partialmethod(logger.__class__.log, "INIT_WARN")
logger.__class__.init_err = partialmethod(logger.__class__.log, "INIT_ERR")
logger.__class__.message = partialmethod(logger.__class__.log, "MESSAGE")
logger.__class__.stats = partialmethod(logger.__class__.log, "STATS")

config = {
    "handlers": [
        {
            "sink": sys.stderr,
            "format": logfmt,
            "colorize": True,
            "filter": is_stderr_log,
        },
        {
            "sink": sys.stdout,
            "format": genfmt,
            "level": "PROMPT",
            "colorize": True,
            "filter": is_stdout_log,
        },
        {
            "sink": sys.stdout,
            "format": initfmt,
            "level": "INIT",
            "colorize": True,
            "filter": is_init_log,
        },
        {
            "sink": sys.stdout,
            "format": msgfmt,
            "level": "MESSAGE",
            "colorize": True,
            "filter": is_msg_log,
        },
        {
            "sink": "logs/bridge.log",
            "format": logfmt,
            "level": "DEBUG",
            "colorize": False,
            "filter": is_not_stats_log,
            "retention": "2 days",
            "rotation": "3 hours",
        },
        {
            "sink": "logs/stats.log",
            "format": logfmt,
            "level": "STATS",
            "colorize": False,
            "filter": is_stats_log,
            "retention": "7 days",
            "rotation": "1 days",
        },
        {
            "sink": "logs/trace.log",
            "format": logfmt,
            "level": "TRACE",
            "colorize": False,
            "filter": is_trace_log,
            "retention": "3 days",
            "rotation": "1 days",
            "backtrace": True,
            "diagnose": True,
        },
    ],
}
logger.configure(**config)



import cv2
import numpy as np

from .matlab_cp2tform import get_similarity_transform_for_cv2

# reference facial points, a list of coordinates (x,y)
REFERENCE_FACIAL_POINTS = [
    [30.29459953, 51.69630051],
    [65.53179932, 51.50139999],
    [48.02519989, 71.73660278],
    [33.54930115, 92.3655014],
    [62.72990036, 92.20410156],
]

DEFAULT_CROP_SIZE = (96, 112)


class FaceWarpException(Exception):
    def __str__(self):
        return "In File {}:{}".format(__file__, super.__str__(self))


def get_reference_facial_points(
    output_size=None, inner_padding_factor=0.0, outer_padding=(0, 0), default_square=False
):
    """
    Function:
    ----------
        get reference 5 key points according to crop settings:
        0. Set default crop_size:
            if default_square:
                crop_size = (112, 112)
            else:
                crop_size = (96, 112)
        1. Pad the crop_size by inner_padding_factor in each side;
        2. Resize crop_size into (output_size - outer_padding*2),
            pad into output_size with outer_padding;
        3. Output reference_5point;
    Parameters:
    ----------
        @output_size: (w, h) or None
            size of aligned face image
        @inner_padding_factor: (w_factor, h_factor)
            padding factor for inner (w, h)
        @outer_padding: (w_pad, h_pad)
            each row is a pair of coordinates (x, y)
        @default_square: True or False
            if True:
                default crop_size = (112, 112)
            else:
                default crop_size = (96, 112);
        !!! make sure, if output_size is not None:
                (output_size - outer_padding)
                = some_scale * (default crop_size * (1.0 +
                inner_padding_factor))
    Returns:
    ----------
        @reference_5point: 5x2 np.array
            each row is a pair of transformed coordinates (x, y)
    """

    tmp_5pts = np.array(REFERENCE_FACIAL_POINTS)
    tmp_crop_size = np.array(DEFAULT_CROP_SIZE)

    # 0) make the inner region a square
    if default_square:
        size_diff = max(tmp_crop_size) - tmp_crop_size
        tmp_5pts += size_diff / 2
        tmp_crop_size += size_diff

    if output_size and output_size[0] == tmp_crop_size[0] and output_size[1] == tmp_crop_size[1]:
        return tmp_5pts

    if inner_padding_factor == 0 and outer_padding == (0, 0):
        if output_size is None:
            return tmp_5pts
        else:
            raise FaceWarpException("No paddings to do, output_size must be None or {}".format(tmp_crop_size))

    # check output size
    if not (0 <= inner_padding_factor <= 1.0):
        raise FaceWarpException("Not (0 <= inner_padding_factor <= 1.0)")

    if (inner_padding_factor > 0 or outer_padding[0] > 0 or outer_padding[1] > 0) and output_size is None:
        output_size = tmp_crop_size * (1 + inner_padding_factor * 2).astype(np.int32)
        output_size += np.array(outer_padding)
    if not (outer_padding[0] < output_size[0] and outer_padding[1] < output_size[1]):
        raise FaceWarpException("Not (outer_padding[0] < output_size[0] and outer_padding[1] < output_size[1])")

    # 1) pad the inner region according inner_padding_factor
    if inner_padding_factor > 0:
        size_diff = tmp_crop_size * inner_padding_factor * 2
        tmp_5pts += size_diff / 2
        tmp_crop_size += np.round(size_diff).astype(np.int32)

    # 2) resize the padded inner region
    size_bf_outer_pad = np.array(output_size) - np.array(outer_padding) * 2

    if size_bf_outer_pad[0] * tmp_crop_size[1] != size_bf_outer_pad[1] * tmp_crop_size[0]:
        raise FaceWarpException(
            "Must have (output_size - outer_padding)" "= some_scale * (crop_size * (1.0 + inner_padding_factor)"
        )

    scale_factor = size_bf_outer_pad[0].astype(np.float32) / tmp_crop_size[0]
    tmp_5pts = tmp_5pts * scale_factor
    #    size_diff = tmp_crop_size * (scale_factor - min(scale_factor))
    #    tmp_5pts = tmp_5pts + size_diff / 2
    tmp_crop_size = size_bf_outer_pad

    # 3) add outer_padding to make output_size
    reference_5point = tmp_5pts + np.array(outer_padding)
    tmp_crop_size = output_size

    return reference_5point


def get_affine_transform_matrix(src_pts, dst_pts):
    """
    Function:
    ----------
        get affine transform matrix 'tfm' from src_pts to dst_pts
    Parameters:
    ----------
        @src_pts: Kx2 np.array
            source points matrix, each row is a pair of coordinates (x, y)
        @dst_pts: Kx2 np.array
            destination points matrix, each row is a pair of coordinates (x, y)
    Returns:
    ----------
        @tfm: 2x3 np.array
            transform matrix from src_pts to dst_pts
    """

    tfm = np.float32([[1, 0, 0], [0, 1, 0]])
    n_pts = src_pts.shape[0]
    ones = np.ones((n_pts, 1), src_pts.dtype)
    src_pts_ = np.hstack([src_pts, ones])
    dst_pts_ = np.hstack([dst_pts, ones])

    A, res, rank, s = np.linalg.lstsq(src_pts_, dst_pts_)

    if rank == 3:
        tfm = np.float32([[A[0, 0], A[1, 0], A[2, 0]], [A[0, 1], A[1, 1], A[2, 1]]])
    elif rank == 2:
        tfm = np.float32([[A[0, 0], A[1, 0], 0], [A[0, 1], A[1, 1], 0]])

    return tfm


def warp_and_crop_face(src_img, facial_pts, reference_pts=None, crop_size=(96, 112), align_type="smilarity"):
    """
    Function:
    ----------
        apply affine transform 'trans' to uv
    Parameters:
    ----------
        @src_img: 3x3 np.array
            input image
        @facial_pts: could be
            1)a list of K coordinates (x,y)
        or
            2) Kx2 or 2xK np.array
            each row or col is a pair of coordinates (x, y)
        @reference_pts: could be
            1) a list of K coordinates (x,y)
        or
            2) Kx2 or 2xK np.array
            each row or col is a pair of coordinates (x, y)
        or
            3) None
            if None, use default reference facial points
        @crop_size: (w, h)
            output face image size
        @align_type: transform type, could be one of
            1) 'similarity': use similarity transform
            2) 'cv2_affine': use the first 3 points to do affine transform,
                    by calling cv2.getAffineTransform()
            3) 'affine': use all points to do affine transform
    Returns:
    ----------
        @face_img: output face image with size (w, h) = @crop_size
    """

    if reference_pts is None:
        if crop_size[0] == 96 and crop_size[1] == 112:
            reference_pts = REFERENCE_FACIAL_POINTS
        else:
            default_square = False
            inner_padding_factor = 0
            outer_padding = (0, 0)
            output_size = crop_size

            reference_pts = get_reference_facial_points(
                output_size, inner_padding_factor, outer_padding, default_square
            )

    ref_pts = np.float32(reference_pts)
    ref_pts_shp = ref_pts.shape
    if max(ref_pts_shp) < 3 or min(ref_pts_shp) != 2:
        raise FaceWarpException("reference_pts.shape must be (K,2) or (2,K) and K>2")

    if ref_pts_shp[0] == 2:
        ref_pts = ref_pts.T

    src_pts = np.float32(facial_pts)
    src_pts_shp = src_pts.shape
    if max(src_pts_shp) < 3 or min(src_pts_shp) != 2:
        raise FaceWarpException("facial_pts.shape must be (K,2) or (2,K) and K>2")

    if src_pts_shp[0] == 2:
        src_pts = src_pts.T

    if src_pts.shape != ref_pts.shape:
        raise FaceWarpException("facial_pts and reference_pts must have the same shape")

    if align_type == "cv2_affine":
        tfm = cv2.getAffineTransform(src_pts[0:3], ref_pts[0:3])
    elif align_type == "affine":
        tfm = get_affine_transform_matrix(src_pts, ref_pts)
    else:
        tfm = get_similarity_transform_for_cv2(src_pts, ref_pts)

    face_img = cv2.warpAffine(src_img, tfm, (crop_size[0], crop_size[1]))

    return face_img



import gradio as gr

from nataili.train.dreambooth import DreamboothLoRA

trainer = DreamboothLoRA()


def train(
    base_checkpoint, data_root, project_name, output_dir, max_train_steps, progress=gr.Progress(track_tqdm=True)
):
    trainer.train(
        base_checkpoint=base_checkpoint,
        data_root=data_root,
        project_name=project_name,
        output_dir=output_dir,
        max_train_steps=int(max_train_steps),
        progress_bar=progress,
    )


with gr.Blocks(analytics_enabled=False) as blocks:
    with gr.Row():
        with gr.Column(scale=4):
            base_checkpoint = gr.Textbox(
                label="Base checkpoint",
                placeholder="runwayml/stable-diffusion-v1-5",
                value="runwayml/stable-diffusion-v1-5",
            )
            data_root = gr.Textbox(
                label="Data root",
                placeholder="X:/lora/squishmallow",
            )
            project_name = gr.Textbox(
                label="Project name",
                placeholder="squishmallow",
            )
            output_dir = gr.Textbox(
                label="Output directory",
                placeholder="X:/lora/output",
            )
            max_train_steps = gr.Textbox(
                label="Max train steps",
                placeholder="1000",
            )
            btn = gr.Button("Train")
        with gr.Column(scale=1):
            output_file = gr.Textbox(
                label="Output file",
            )

    btn.click(
        inputs=[
            base_checkpoint,
            data_root,
            project_name,
            output_dir,
            max_train_steps,
        ],
        outputs=[output_file],
        fn=train,
    )

blocks.queue().launch(show_error=True, quiet=False, debug=True)



import os

from PIL import Image

from nataili.model_manager.clip import ClipModelManager
from nataili.clip.interrogate import Interrogator
from nataili.util.logger import logger

images = []

directory = "teeth"

mm = ClipModelManager()

mm.load("ViT-L/14")

interrogator = Interrogator(
    mm.loaded_models["ViT-L/14"],
)

base_words = [
    'teeth',
    # 'hand',
    # 'mouth',
    # 'fingers',
]

description_words = [
    'ugly',
    # 'bad',
    # 'misshaped',
    # 'misfigured',
    # 'malformed',
    # 'deformed',
    # 'disfigured',
    # 'disabled',
    # 'good',
    # 'beautiful',
    # 'perfect',
    # 'well-formed',
    # 'nice',
    'pretty',
]

word_list = []

for base_word in base_words:
    for description_word in description_words:
        word_list.append(f"{description_word} {base_word}")
    word_list.append(base_word) 

html_string = ""



for file in os.listdir(directory):
    results = interrogator(filename=file, directory=directory, text_array=word_list, similarity=True)
    results = results['default']
    html_string += f"""
    <h1>{file}</h1>
    <img src="{directory}/{file}" width="300" />
    <table>
    <tr>
    <th>Word</th>
    <th>Score</th>
    </tr>
    """
    for key in results.keys():
        html_string += f"""
        <tr>
        <td>{key}</td>
        <td>{results[key]}</td>
        </tr>
        """
    html_string += "</table>"
    html_string += "<hr />"

with open("test.html", "w") as f:
    f.write(html_string)



import math
import os
import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint

from transformers import T5Tokenizer, T5EncoderModel, CLIPTokenizer, CLIPTextModel

import open_clip
from ldm2.util import default, count_params
from nataili.util.logger import logger


class AbstractEncoder(nn.Module):
    def __init__(self):
        super().__init__()

    def encode(self, *args, **kwargs):
        raise NotImplementedError


class IdentityEncoder(AbstractEncoder):

    def encode(self, x):
        return x


class ClassEmbedder(nn.Module):
    def __init__(self, embed_dim, n_classes=1000, key='class', ucg_rate=0.1):
        super().__init__()
        self.key = key
        self.embedding = nn.Embedding(n_classes, embed_dim)
        self.n_classes = n_classes
        self.ucg_rate = ucg_rate

    def forward(self, batch, key=None, disable_dropout=False):
        if key is None:
            key = self.key
        # this is for use in crossattn
        c = batch[key][:, None]
        if self.ucg_rate > 0. and not disable_dropout:
            mask = 1. - torch.bernoulli(torch.ones_like(c) * self.ucg_rate)
            c = mask * c + (1-mask) * torch.ones_like(c)*(self.n_classes-1)
            c = c.long()
        c = self.embedding(c)
        return c

    def get_unconditional_conditioning(self, bs, device="cuda"):
        uc_class = self.n_classes - 1  # 1000 classes --> 0 ... 999, one extra class for ucg (class 1000)
        uc = torch.ones((bs,), device=device) * uc_class
        uc = {self.key: uc}
        return uc


def disabled_train(self, mode=True):
    """Overwrite model.train with this function to make sure train/eval mode
    does not change anymore."""
    return self


class FrozenT5Embedder(AbstractEncoder):
    """Uses the T5 transformer encoder for text"""
    def __init__(self, version="google/t5-v1_1-large", device="cuda", max_length=77, freeze=True):  # others are google/t5-v1_1-xl and google/t5-v1_1-xxl
        super().__init__()
        self.tokenizer = T5Tokenizer.from_pretrained(version)
        self.transformer = T5EncoderModel.from_pretrained(version)
        self.device = device
        self.max_length = max_length   # TODO: typical value?
        if freeze:
            self.freeze()

    def freeze(self):
        self.transformer = self.transformer.eval()
        #self.train = disabled_train
        for param in self.parameters():
            param.requires_grad = False

    def forward(self, text):
        batch_encoding = self.tokenizer(text, truncation=True, max_length=self.max_length, return_length=True,
                                        return_overflowing_tokens=False, padding="max_length", return_tensors="pt")
        tokens = batch_encoding["input_ids"].to(self.device)
        outputs = self.transformer(input_ids=tokens)

        z = outputs.last_hidden_state
        return z

    def encode(self, text):
        return self(text)


class FrozenCLIPEmbedder(AbstractEncoder):
    """Uses the CLIP transformer encoder for text (from Hugging Face)"""
    def __init__(self, version="openai/clip-vit-large-patch14", device="cuda", max_length=77):
        super().__init__()
        # TODO: Update this
        if os.path.exists("models/clip-vit-large-patch14"):
            self.tokenizer = CLIPTokenizer.from_pretrained("models/clip-vit-large-patch14")
            self.transformer = CLIPTextModel.from_pretrained("models/clip-vit-large-patch14")
        else:
            self.tokenizer = CLIPTokenizer.from_pretrained(version)
            self.transformer = CLIPTextModel.from_pretrained(version)
        self.device = device
        self.max_length = max_length
        self.chunk_length = 75
        self.freeze()

        vocab = self.tokenizer.get_vocab()
        self.comma_token = vocab.get(',</w>', None)
        self.id_start = self.tokenizer.bos_token_id
        self.id_end = self.tokenizer.eos_token_id
        self.id_pad = self.id_end

    def empty_chunk(self):
        """creates an empty PromptChunk and returns it"""

        chunk = PromptChunk()
        chunk.tokens = [self.id_start] + [self.id_end] * (self.chunk_length + 1)
        return chunk

    def freeze(self):
        self.transformer = self.transformer.eval()
        for param in self.parameters():
            param.requires_grad = False

    def get_target_prompt_token_count(self, token_count):
        """returns the maximum number of tokens a prompt of a known length can have before it requires one more PromptChunk to be represented"""

        return math.ceil(max(token_count, 1) / self.chunk_length) * self.chunk_length

    def tokenize(self, texts):
        tokenized = self.tokenizer(texts, truncation=False, add_special_tokens=False)["input_ids"]

        return tokenized

    def encode_with_transformers(self, tokens, clip_skip):
        outputs = self.transformer(input_ids=tokens, output_hidden_states=True)

        if clip_skip > 1:
            z = outputs.hidden_states[-clip_skip]
            z = self.transformer.text_model.final_layer_norm(z)
        else:
            z = outputs.last_hidden_state

        return z

    def encode_embedding_init_text(self, init_text, nvpt):
        """Converts text into a tensor with this text's tokens' embeddings. Note that those are embeddings before they are passed through
        transformers. nvpt is used as a maximum length in tokens. If text produces less teokens than nvpt, only this many is returned."""

        raise NotImplementedError
    
    def tokenize_line(self, line):
        """
        this transforms a single prompt into a list of PromptChunk objects - as many as needed to
        represent the prompt.
        Returns the list and the total number of tokens in the prompt.
        """
        tokenized = self.tokenize(line)

        chunks = []
        chunk = PromptChunk()
        token_count = 0
        last_comma = -1

        def next_chunk():
            """puts current chunk into the list of results and produces the next one - empty"""
            nonlocal token_count
            nonlocal last_comma
            nonlocal chunk
            token_count += len(chunk.tokens)
            to_add = self.chunk_length - len(chunk.tokens)
            if to_add > 0:
                chunk.tokens += [self.id_end] * to_add

            chunk.tokens = [self.id_start] + chunk.tokens + [self.id_end]

            last_comma = -1
            chunks.append(chunk)
            chunk = PromptChunk()

        for tokens in zip(tokenized):
            position = 0
            while position < len(tokens):
                token = tokens[position]

                if token == self.comma_token:
                    last_comma = len(chunk.tokens)

                elif len(chunk.tokens) == self.chunk_length and last_comma != -1 and len(chunk.tokens) - last_comma <= 20:
                    break_location = last_comma + 1

                    reloc_tokens = chunk.tokens[break_location:]
                    chunk.tokens = chunk.tokens[:break_location]

                    next_chunk()
                    chunk.tokens = reloc_tokens

                if len(chunk.tokens) == self.chunk_length:
                    next_chunk()

                chunk.tokens.append(token)
                position += 1

        if len(chunk.tokens) > 0 or len(chunks) == 0:
            next_chunk()

        return chunks, token_count

    def process_texts(self, text):
        token_count = 0
        batch_chunks = []
        chunks, current_token_count = self.tokenize_line(text)
        token_count = max(current_token_count, token_count)

        batch_chunks.append(chunks)
        return batch_chunks, token_count
    
    def process_tokens(self, remade_batch_tokens, clip_skip):
        """
        sends one single prompt chunk to be encoded by transformers neural network.
        remade_batch_tokens is a batch of tokens - a list, where every element is a list of tokens; usually
        there are exactly 77 tokens in the list.
        """
        tokens = torch.asarray(remade_batch_tokens).to(self.device)

        # this is for SD2: SD1 uses the same token for padding and end of text, while SD2 uses different ones.
        if self.id_end != self.id_pad:
            for batch_pos in range(len(remade_batch_tokens)):
                index = remade_batch_tokens[batch_pos].index(self.id_end)
                tokens[batch_pos, index+1:tokens.shape[1]] = self.id_pad

        z = self.encode_with_transformers(tokens, clip_skip)

        return z

    def forward(self, text, clip_skip):
        if len(text) < 50:
            batch_encoding = self.tokenizer(text, truncation=True, max_length=self.max_length, return_length=True,
                                            return_overflowing_tokens=False, padding="max_length", return_tensors="pt")
            tokens = batch_encoding["input_ids"].to(self.device)
            z = self.encode_with_transformers(tokens, clip_skip)

            return z
        
        else:
            batch_chunks, token_count = self.process_texts(text)
            logger.debug(f"Token count for prompt = {token_count}")
            chunk_count = max([len(x) for x in batch_chunks])

            zs = []
            for i in range(chunk_count):
                batch_chunk = [chunks[i] if i < len(chunks) else self.empty_chunk() for chunks in batch_chunks]

                tokens = [x.tokens for x in batch_chunk]

                z = self.process_tokens(tokens, clip_skip)
                zs.append(z)

            return torch.hstack(zs)

    def encode(self, text, clip_skip):
        return self(text, clip_skip)


class FrozenOpenCLIPEmbedder(AbstractEncoder):
    """
    Uses the OpenCLIP transformer encoder for text
    """
    LAYERS = [
        #"pooled",
        "last",
        "penultimate"
    ]
    def __init__(self, arch="ViT-H-14", version="laion2b_s32b_b79k", device="cuda", max_length=77,
                 freeze=True, layer="last"):
        super().__init__()
        assert layer in self.LAYERS
        model, _, _ = open_clip.create_model_and_transforms(arch, device=torch.device('cpu'), pretrained=version)
        del model.visual
        self.model = model

        self.device = device
        self.max_length = max_length
        if freeze:
            self.freeze()
        self.layer = layer
        if self.layer == "last":
            self.layer_idx = 0
        elif self.layer == "penultimate":
            self.layer_idx = 1
        else:
            raise NotImplementedError()

    def freeze(self):
        self.model = self.model.eval()
        for param in self.parameters():
            param.requires_grad = False

    def forward(self, text, clip_skip=None):
        tokens = open_clip.tokenize(text)
        z = self.encode_with_transformer(tokens.to(self.device))
        return z

    def encode_with_transformer(self, text):
        x = self.model.token_embedding(text)  # [batch_size, n_ctx, d_model]
        x = x + self.model.positional_embedding
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.text_transformer_forward(x, attn_mask=self.model.attn_mask)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.model.ln_final(x)
        return x

    def text_transformer_forward(self, x: torch.Tensor, attn_mask = None):
        for i, r in enumerate(self.model.transformer.resblocks):
            if i == len(self.model.transformer.resblocks) - self.layer_idx:
                break
            if self.model.transformer.grad_checkpointing and not torch.jit.is_scripting():
                x = checkpoint(r, x, attn_mask)
            else:
                x = r(x, attn_mask=attn_mask)
        return x

    def encode(self, text, clip_skip=None):
        return self(text)


class FrozenCLIPT5Encoder(AbstractEncoder):
    def __init__(self, clip_version="openai/clip-vit-large-patch14", t5_version="google/t5-v1_1-xl", device="cuda",
                 clip_max_length=77, t5_max_length=77):
        super().__init__()
        self.clip_encoder = FrozenCLIPEmbedder(clip_version, device, max_length=clip_max_length)
        self.t5_encoder = FrozenT5Embedder(t5_version, device, max_length=t5_max_length)
        print(f"{self.clip_encoder.__class__.__name__} has {count_params(self.clip_encoder)*1.e-6:.2f} M parameters, "
              f"{self.t5_encoder.__class__.__name__} comes with {count_params(self.t5_encoder)*1.e-6:.2f} M params.")

    def encode(self, text):
        return self(text)

    def forward(self, text):
        clip_z = self.clip_encoder.encode(text)
        t5_z = self.t5_encoder.encode(text)
        return [clip_z, t5_z]


class PromptChunk:

    def __init__(self):
        self.tokens = []



import os

from annotator.uniformer.mmseg.apis import init_segmentor, inference_segmentor, show_result_pyplot
from annotator.uniformer.mmseg.core.evaluation import get_palette
from annotator.util import annotator_ckpts_path, nataili_package


checkpoint_file = "https://huggingface.co/lllyasviel/ControlNet/resolve/main/annotator/ckpts/upernet_global_small.pth"


class UniformerDetector:
    def __init__(self):
        modelpath = os.path.join(annotator_ckpts_path, "upernet_global_small.pth")
        if not os.path.exists(modelpath):
            from basicsr.utils.download_util import load_file_from_url
            load_file_from_url(checkpoint_file, model_dir=annotator_ckpts_path)
        config_file = str(nataili_package / "uniformer"/"exp"/"upernet_global_small"/"config.py")
        self.model = init_segmentor(config_file, modelpath).cuda()

    def __call__(self, img):
        result = inference_segmentor(self.model, img)
        res_img = show_result_pyplot(self.model, img, result, get_palette('ade'), opacity=1)
        del self.model
        return res_img



# Copyright (c) OpenMMLab. All rights reserved.
from .base import LoggerHook
from .dvclive import DvcliveLoggerHook
from .mlflow import MlflowLoggerHook
from .neptune import NeptuneLoggerHook
from .pavi import PaviLoggerHook
from .tensorboard import TensorboardLoggerHook
from .text import TextLoggerHook
from .wandb import WandbLoggerHook

__all__ = [
    'LoggerHook', 'MlflowLoggerHook', 'PaviLoggerHook',
    'TensorboardLoggerHook', 'TextLoggerHook', 'WandbLoggerHook',
    'NeptuneLoggerHook', 'DvcliveLoggerHook'
]



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import time
from pathlib import Path

import clip
import open_clip
import torch

from nataili.cache import get_cache_directory
from nataili.model_manager.base import BaseModelManager
from nataili.util.load_list import load_list
from nataili.util.logger import logger


class ClipModelManager(BaseModelManager):
    def __init__(self, download_reference=True):
        super().__init__()
        self.download_reference = download_reference
        self.path = f"{get_cache_directory()}/clip"
        self.models_db_name = "clip"
        self.models_path = self.pkg / f"{self.models_db_name}.json"
        self.remote_db = (
            f"https://raw.githubusercontent.com/db0/AI-Horde-image-model-reference/main/{self.models_db_name}.json"
        )
        self.init(list_models=True)

    def load_data_lists(self):
        data_lists = {}
        data_lists["artist"] = load_list(self.pkg / "artists.txt")
        data_lists["flavors"] = load_list(self.pkg / "flavors.txt")
        data_lists["medium"] = load_list(self.pkg / "mediums.txt")
        data_lists["movement"] = load_list(self.pkg / "movements.txt")
        data_lists["trending"] = load_list(self.pkg / "sites.txt")
        data_lists["techniques"] = load_list(self.pkg / "techniques.txt")
        data_lists["tags"] = load_list(self.pkg / "tags.txt")
        return data_lists

    def load_coca(self, model_name, half_precision=True, gpu_id=0, cpu_only=False):
        model_path = self.get_model_files(model_name)[0]["path"]
        model_path = f"{self.path}/{model_path}"
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        model, _, transform = open_clip.create_model_and_transforms(
            "coca_ViT-L-14",
            pretrained=model_path,
            device=device,
            precision="fp16" if half_precision else "fp32",
        )
        model = model.eval()
        model.to(device)
        if half_precision:
            model = model.half()
        return {
            "model": model,
            "device": device,
            "transform": transform,
            "half_precision": half_precision,
            "cache_name": model_name.replace("/", "_"),
        }

    def load_open_clip(self, model_name, half_precision=True, gpu_id=0, cpu_only=False):
        pretrained = self.get_model(model_name)["pretrained_name"]
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        model, _, preprocess = open_clip.create_model_and_transforms(
            model_name,
            pretrained=pretrained,
            cache_dir=self.path,
            device=device,
            precision="fp16" if half_precision else "fp32",
        )
        model = model.eval()
        model.to(device)
        if half_precision:
            model = model.half()
        data_lists = self.load_data_lists()
        return {
            "model": model,
            "device": device,
            "preprocess": preprocess,
            "data_lists": data_lists,
            "half_precision": half_precision,
            "cache_name": model_name.replace("/", "_"),
        }

    def load_clip(self, model_name, half_precision=True, gpu_id=0, cpu_only=False):
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        model, preprocess = clip.load(model_name, device=device, download_root=self.path)
        model = model.eval()
        if half_precision:
            model = model.half()
        data_lists = self.load_data_lists()
        return {
            "model": model,
            "device": device,
            "preprocess": preprocess,
            "data_lists": data_lists,
            "half_precision": half_precision,
            "cache_name": model_name.replace("/", "_"),
        }

    def load(self, model_name: str, half_precision=True, gpu_id=0, cpu_only=False):
        """
        model_name: str. Name of the model to load. See available_models for a list of available models.
        half_precision: bool. If True, the model will be loaded in half precision.
        gpu_id: int. The id of the gpu to use. If the gpu is not available, the model will be loaded on the cpu.
        cpu_only: bool. If True, the model will be loaded on the cpu. If True, half_precision will be set to False.
        """
        if model_name not in self.models:
            logger.error(f"{model_name} not found")
            return False
        if model_name not in self.available_models:
            logger.error(f"{model_name} not available")
            logger.init_ok(f"Downloading {model_name}", status="Downloading")
            self.download_model(model_name)
            logger.init_ok(f"{model_name} downloaded", status="Downloading")
        if model_name not in self.loaded_models:
            if not self.cuda_available:
                cpu_only = True
            tic = time.time()
            logger.init(f"{model_name}", status="Loading")
            if self.models[model_name]["type"] == "open_clip":
                self.loaded_models[model_name] = self.load_open_clip(model_name, half_precision, gpu_id, cpu_only)
            elif self.models[model_name]["type"] == "clip":
                self.loaded_models[model_name] = self.load_clip(model_name, half_precision, gpu_id, cpu_only)
            elif self.models[model_name]["type"] == "coca":
                self.loaded_models[model_name] = self.load_coca(model_name, half_precision, gpu_id, cpu_only)
            else:
                logger.error(f"Unknown model type: {self.models[model_name]['type']}")
                return
            logger.init_ok(f"Loading {model_name}", status="Success")
            toc = time.time()
            logger.init_ok(f"Loading {model_name}: Took {toc-tic} seconds", status="Success")
            return True



import torch


class BaseModel(torch.nn.Module):
    def load(self, path):
        """Load model from file.

        Args:
            path (str): file path
        """
        parameters = torch.load(path, map_location=torch.device('cpu'))

        if "optimizer" in parameters:
            parameters = parameters["model"]

        self.load_state_dict(parameters)



_base_ = [
    '../../configs/_base_/models/upernet_uniformer.py', 
    '../../configs/_base_/datasets/ade20k.py',
    '../../configs/_base_/default_runtime.py', 
    '../../configs/_base_/schedules/schedule_160k.py'
]
model = dict(
    backbone=dict(
        type='UniFormer',
        embed_dim=[64, 128, 320, 512],
        layers=[3, 4, 8, 3],
        head_dim=64,
        drop_path_rate=0.25,
        windows=True,
        hybrid=False,
        window_size=32
    ),
    decode_head=dict(
        in_channels=[64, 128, 320, 512],
        num_classes=150
    ),
    auxiliary_head=dict(
        in_channels=320,
        num_classes=150
    ))

# AdamW optimizer, no weight decay for position embedding & layer norm in backbone
optimizer = dict(_delete_=True, type='AdamW', lr=0.00006, betas=(0.9, 0.999), weight_decay=0.01,
                 paramwise_cfg=dict(custom_keys={'absolute_pos_embed': dict(decay_mult=0.),
                                                 'relative_position_bias_table': dict(decay_mult=0.),
                                                 'norm': dict(decay_mult=0.)}))

lr_config = dict(_delete_=True, policy='poly',
                 warmup='linear',
                 warmup_iters=1500,
                 warmup_ratio=1e-6,
                 power=1.0, min_lr=0.0, by_epoch=False)

data=dict(samples_per_gpu=2)


import torch


def realesrgan(bg_tile):
    from basicsr.archs.rrdbnet_arch import RRDBNet
    from realesrgan import RealESRGANer

    cuda_is_available = torch.cuda.is_available()
    half = True if cuda_is_available else False
    model = RRDBNet(
        num_in_ch=3,
        num_out_ch=3,
        num_feat=64,
        num_block=23,
        num_grow_ch=32,
        scale=2,
    )
    upsampler = RealESRGANer(
        scale=2,
        model_path="https://github.com/sczhou/CodeFormer/releases/download/v0.1.0/RealESRGAN_x2plus.pth",
        model=model,
        tile=bg_tile,
        tile_pad=40,
        pre_pad=0,
        half=half,  # need to set False in CPU mode
    )

    if not cuda_is_available:  # CPU
        import warnings

        warnings.warn(
            "Running on CPU now! Make sure your PyTorch version matches your CUDA."
            "The unoptimized RealESRGAN is slow on CPU. "
            "If you want to disable it, please remove `--bg_upsampler` and `--face_upsample` in command.",
            category=RuntimeWarning,
        )
    return upsampler



from ldm2.modules.image_degradation.bsrgan import degradation_bsrgan_variant as degradation_fn_bsr
from ldm2.modules.image_degradation.bsrgan_light import degradation_bsrgan_variant as degradation_fn_bsr_light



#  Copyright (c) Meta Platforms, Inc. and affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Functions for working with torch Tensors.
AITemplate doesn't depend on PyTorch, but it exposes
many APIs that work with torch Tensors anyways.

The functions in this file may assume that
`import torch` will work.
"""


def torch_dtype_to_string(dtype):
    import torch

    dtype_to_str = {
        torch.float16: "float16",
        torch.float32: "float32",
        torch.int32: "int32",
        torch.int64: "int64",
        torch.bool: "bool",
    }
    if dtype not in dtype_to_str:
        raise ValueError(f"Got unsupported input dtype {dtype}! Supported dtypes are: {list(dtype_to_str.keys())}")
    return dtype_to_str[dtype]



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import re
from contextlib import nullcontext

import PIL.ImageOps
import torch
from PIL import Image
from slugify import slugify

from nataili import disable_voodoo
from nataili.util.cache import torch_gc
from nataili.util.cast import autocast_cuda
from nataili.util.get_next_sequence_number import get_next_sequence_number
from nataili.util.logger import logger
from nataili.util.process_prompt_tokens import process_prompt_tokens
from nataili.util.save_sample import save_sample
from nataili.util.seed_to_int import seed_to_int

try:
    from nataili.util.voodoo import load_diffusers_pipeline_from_plasma
except ModuleNotFoundError as e:
    if not disable_voodoo.active:
        raise e


class inpainting:
    def __init__(
        self,
        pipe,
        device,
        output_dir,
        save_extension="jpg",
        output_file_path=False,
        load_concepts=False,
        concepts_dir=None,
        verify_input=True,
        auto_cast=True,
        filter_nsfw=False,
        disable_voodoo=False,
    ):
        self.output_dir = output_dir
        self.output_file_path = output_file_path
        self.save_extension = save_extension
        self.load_concepts = load_concepts
        self.concepts_dir = concepts_dir
        self.verify_input = verify_input
        self.auto_cast = auto_cast
        self.pipe = pipe
        self.device = device
        self.comments = []
        self.output_images = []
        self.info = ""
        self.stats = ""
        self.images = []
        self.filter_nsfw = filter_nsfw
        self.disable_voodoo = disable_voodoo

    def resize_image(self, resize_mode, im: Image.Image, width, height):
        LANCZOS = Image.Resampling.LANCZOS if hasattr(Image, "Resampling") else Image.LANCZOS
        if resize_mode == "resize":
            res = im.resize((width, height), resample=LANCZOS)
        elif resize_mode == "crop":
            ratio = width / height
            src_ratio = im.width / im.height

            src_w = width if ratio > src_ratio else im.width * height // im.height
            src_h = height if ratio <= src_ratio else im.height * width // im.width

            resized = im.resize((src_w, src_h), resample=LANCZOS)
            res = Image.new("RGBA", (width, height))
            res.paste(resized, box=(width // 2 - src_w // 2, height // 2 - src_h // 2))
        else:
            ratio = width / height
            src_ratio = im.width / im.height

            src_w = width if ratio < src_ratio else im.width * height // im.height
            src_h = height if ratio >= src_ratio else im.height * width // im.width

            resized = im.resize((src_w, src_h), resample=LANCZOS)
            res = Image.new("RGBA", (width, height))
            res.paste(resized, box=(width // 2 - src_w // 2, height // 2 - src_h // 2))

            if ratio < src_ratio:
                fill_height = height // 2 - src_h // 2
                res.paste(
                    resized.resize((width, fill_height), box=(0, 0, width, 0)),
                    box=(0, 0),
                )
                res.paste(
                    resized.resize(
                        (width, fill_height),
                        box=(0, resized.height, width, resized.height),
                    ),
                    box=(0, fill_height + src_h),
                )
            elif ratio > src_ratio:
                fill_width = width // 2 - src_w // 2
                res.paste(
                    resized.resize((fill_width, height), box=(0, 0, 0, height)),
                    box=(0, 0),
                )
                res.paste(
                    resized.resize(
                        (fill_width, height),
                        box=(resized.width, 0, resized.width, height),
                    ),
                    box=(fill_width + src_w, 0),
                )

        return res

    @autocast_cuda
    def generate(
        self,
        prompt: str,
        inpaint_img=None,
        inpaint_mask=None,
        ddim_steps=50,
        n_iter=1,
        batch_size=1,
        cfg_scale=7.5,
        seed=None,
        height=512,
        width=512,
        save_individual_images: bool = True,
    ):
        seed = seed_to_int(seed)
        inpaint_img = self.resize_image("resize", inpaint_img, width, height)

        # mask information has been transferred in the Alpha channel of the inpaint image
        logger.debug(inpaint_mask)
        if inpaint_mask is None:
            try:
                red, green, blue, alpha = inpaint_img.split()
            except ValueError:
                raise Exception("inpainting image doesn't have an alpha channel.")

            inpaint_mask = alpha
            inpaint_mask = PIL.ImageOps.invert(inpaint_mask)
        else:
            inpaint_mask = self.resize_image("resize", inpaint_mask, width, height)

        torch_gc()

        # TODO: add diffusers support to process_prompt_tokens
        # if self.load_concepts and self.concepts_dir is not None:
        #     prompt_tokens = re.findall("<([a-zA-Z0-9-]+)>", prompt)
        #     if prompt_tokens:
        #         process_prompt_tokens(prompt_tokens)

        os.makedirs(self.output_dir, exist_ok=True)

        sample_path = os.path.join(self.output_dir, "samples")
        os.makedirs(sample_path, exist_ok=True)

        all_prompts = batch_size * [prompt]
        all_seeds = [seed + x for x in range(len(all_prompts))]

        precision_scope = torch.autocast if self.auto_cast else nullcontext

        with torch.no_grad(), precision_scope("cuda"):
            for n in range(batch_size):
                logger.debug(f"Iteration: {n+1}/{batch_size}")

                prompt = all_prompts[n]
                seed = all_seeds[n]
                # print("prompt: " + prompt + ", seed: " + str(seed))

                generator = torch.Generator(device=self.device).manual_seed(seed)

                if not self.disable_voodoo:
                    with load_diffusers_pipeline_from_plasma(self.pipe, self.device) as pipe:
                        safety_checker = None
                        if not self.filter_nsfw:
                            safety_checker = pipe.safety_checker
                            pipe.safety_checker = None

                        x_samples = pipe(
                            prompt=prompt,
                            image=inpaint_img,
                            mask_image=inpaint_mask,
                            guidance_scale=cfg_scale,
                            num_inference_steps=ddim_steps,
                            generator=generator,
                            num_images_per_prompt=n_iter,
                            width=width,
                            height=height,
                        ).images

                        if safety_checker:
                            pipe.safety_checker = safety_checker
                else:
                    safety_checker = None
                    if not self.filter_nsfw:
                        safety_checker = self.pipe.safety_checker
                        self.pipe.safety_checker = None

                    x_samples = self.pipe(
                        prompt=prompt,
                        image=inpaint_img,
                        mask_image=inpaint_mask,
                        guidance_scale=cfg_scale,
                        num_inference_steps=ddim_steps,
                        generator=generator,
                        num_images_per_prompt=n_iter,
                        width=width,
                        height=height,
                    ).images

                    if safety_checker:
                        self.pipe.safety_checker = safety_checker

                for i, x_sample in enumerate(x_samples):
                    image_dict = {"seed": seed, "image": x_sample}

                    self.images.append(image_dict)

                    if save_individual_images:
                        sanitized_prompt = slugify(prompt)
                        sample_path_i = sample_path
                        base_count = get_next_sequence_number(sample_path_i)
                        full_path = os.path.join(os.getcwd(), sample_path)
                        filename = f"{base_count:05}-{ddim_steps}_{seed}_{sanitized_prompt}"[: 200 - len(full_path)]

                        path = os.path.join(sample_path, filename + "." + self.save_extension)
                        success = save_sample(x_sample, filename, sample_path_i, self.save_extension)

                        if success:
                            if self.output_file_path:
                                self.output_images.append(path)
                            else:
                                self.output_images.append(x_sample)
                        else:
                            return

        self.info = f"""
                {prompt}
                Steps: {ddim_steps}, CFG scale: {cfg_scale}, Seed: {seed}
                """.strip()
        self.stats = """
                """

        for comment in self.comments:
            self.info += "\n\n" + comment

        torch_gc()

        del generator

        return



from .util.switch import Switch

disable_xformers = Switch()
disable_voodoo = Switch()
enable_local_ray_temp = Switch()
disable_progress = Switch()
disable_download_progress = Switch()
enable_ray_alternative = Switch()


class InvalidModelException(Exception):
    pass


class InvalidModelCacheException(InvalidModelException):
    pass



"""
Copyright [2022] Victor C Hall

Licensed under the GNU Affero General Public License;
You may not use this code except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import random
import zipfile

import PIL
import tqdm
from colorama import Fore, Style
from PIL import Image

from nataili import disable_progress

from . import aspects as aspects
from .image_train_item import ImageTrainItem

PIL.Image.MAX_IMAGE_PIXELS = 715827880 * 4  # increase decompression bomb error limit to 4x default


class DataLoaderMultiAspect:
    """
    Data loader for multi-aspect-ratio training and bucketing

    data_root: root folder of training data
    batch_size: number of images per batch
    flip_p: probability of flipping image horizontally (i.e. 0-0.5)
    """

    def __init__(self, data_root, seed=555, debug_level=0, batch_size=1, flip_p=0.0, resolution=512, log_folder=None):
        self.image_paths = []
        self.debug_level = debug_level
        self.flip_p = flip_p
        self.log_folder = log_folder
        self.seed = seed
        self.batch_size = batch_size
        self.runts = []

        self.aspects = aspects.get_aspect_buckets(resolution=resolution, square_only=False)
        logging.info(f"* DLMA resolution {resolution}, buckets: {self.aspects}")
        logging.info(" Preloading images...")

        self.unzip_all(data_root)

        self.__recurse_data_root(self=self, recurse_root=data_root)
        random.Random(seed).shuffle(self.image_paths)
        self.prepared_train_data = self.__prescan_images(self.image_paths, flip_p)  # ImageTrainItem[]
        self.image_caption_pairs = self.__bucketize_images(
            self.prepared_train_data, batch_size=batch_size, debug_level=debug_level
        )

    def shuffle(self):
        self.runts = []
        self.seed = self.seed + 1
        random.Random(self.seed).shuffle(self.prepared_train_data)
        self.image_caption_pairs = self.__bucketize_images(
            self.prepared_train_data, batch_size=self.batch_size, debug_level=0
        )

    def unzip_all(self, path):
        try:
            for root, dirs, files in os.walk(path):
                for file in files:
                    if file.endswith(".zip"):
                        logging.info(f"Unzipping {file}")
                        with zipfile.ZipFile(path, "r") as zip_ref:
                            zip_ref.extractall(path)
        except Exception as e:
            logging.error(f"Error unzipping files {e}")

    def get_all_images(self):
        return self.image_caption_pairs

    @staticmethod
    def __read_caption_from_file(file_path, fallback_caption):
        caption = fallback_caption
        try:
            with open(file_path, encoding="utf-8", mode="r") as caption_file:
                caption = caption_file.read()
        except Exception:
            logging.error(f" *** Error reading {file_path} to get caption, falling back to filename")
            caption = fallback_caption
            pass
        return caption

    def __prescan_images(self, image_paths: list, flip_p=0.0):
        """
        Create ImageTrainItem objects with metadata for hydration later
        """
        decorated_image_train_items = []

        for pathname in tqdm.tqdm(image_paths, disable=disable_progress.active):
            caption_from_filename = os.path.splitext(os.path.basename(pathname))[0].split("_")[0]

            txt_file_path = os.path.splitext(pathname)[0] + ".txt"
            caption_file_path = os.path.splitext(pathname)[0] + ".caption"

            if os.path.exists(txt_file_path):
                caption = self.__read_caption_from_file(txt_file_path, caption_from_filename)
            elif os.path.exists(caption_file_path):
                caption = self.__read_caption_from_file(caption_file_path, caption_from_filename)
            else:
                caption = caption_from_filename

            try:
                image = Image.open(pathname)
                width, height = image.size
                image_aspect = width / height

                target_wh = min(self.aspects, key=lambda aspects: abs(aspects[0] / aspects[1] - image_aspect))

                image_train_item = ImageTrainItem(
                    image=None, caption=caption, target_wh=target_wh, pathname=pathname, flip_p=flip_p
                )

                decorated_image_train_items.append(image_train_item)
            except Exception as e:
                logging.error(
                    f"{Fore.LIGHTRED_EX} *** Error opening {Fore.LIGHTYELLOW_EX}{pathname}{Fore.LIGHTRED_EX} to get metadata. File may be corrupt and will be skipped.{Style.RESET_ALL}"
                )
                logging.error(f" *** exception: {e}")
                pass

        return decorated_image_train_items

    def __bucketize_images(self, prepared_train_data: list, batch_size=1, debug_level=0):
        """
        Put images into buckets based on aspect ratio with batch_size*n images per bucket, discards remainder
        """
        # TODO: this is not terribly efficient but at least linear time
        buckets = {}

        for image_caption_pair in prepared_train_data:
            image_caption_pair.runt_size = 0
            target_wh = image_caption_pair.target_wh

            if (target_wh[0], target_wh[1]) not in buckets:
                buckets[(target_wh[0], target_wh[1])] = []
            buckets[(target_wh[0], target_wh[1])].append(image_caption_pair)

        if len(buckets) > 1:
            for bucket in buckets:
                truncate_count = len(buckets[bucket]) % batch_size
                if truncate_count > 0:
                    runt_bucket = buckets[bucket][-truncate_count:]
                    for item in runt_bucket:
                        item.runt_size = truncate_count
                    while len(runt_bucket) < batch_size:
                        runt_bucket.append(random.choice(runt_bucket))

                    current_bucket_size = len(buckets[bucket])

                    buckets[bucket] = buckets[bucket][: current_bucket_size - truncate_count]
                    buckets[bucket].extend(runt_bucket)

        # flatten the buckets
        image_caption_pairs = []
        for bucket in buckets:
            image_caption_pairs.extend(buckets[bucket])

        return image_caption_pairs

    @staticmethod
    def __recurse_data_root(self, recurse_root):
        multiply = 1
        multiply_path = os.path.join(recurse_root, "multiply.txt")
        if os.path.exists(multiply_path):
            try:
                with open(multiply_path, encoding="utf-8", mode="r") as f:
                    multiply = int(float(f.read().strip()))
                    logging.info(f" * DLMA multiply.txt in {recurse_root} set to {multiply}")
            except Exception:
                logging.error(f" *** Error reading multiply.txt in {recurse_root}, defaulting to 1")
                pass

        for f in os.listdir(recurse_root):
            current = os.path.join(recurse_root, f)

            if os.path.isfile(current):
                ext = os.path.splitext(f)[1]
                if ext.lower() in [".jpg", ".jpeg", ".png", ".bmp", ".webp", ".jfif"]:
                    # add image multiplyrepeats number of times
                    for _ in range(multiply):
                        self.image_paths.append(current)

        sub_dirs = []

        for d in os.listdir(recurse_root):
            current = os.path.join(recurse_root, d)
            if os.path.isdir(current):
                sub_dirs.append(current)

        for dir in sub_dirs:
            self.__recurse_data_root(self=self, recurse_root=dir)



import os

import numpy as np
import torch
import torchvision
from PIL import Image
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.utilities.distributed import rank_zero_only


class ImageLogger(Callback):
    def __init__(self, batch_frequency=2000, max_images=4, clamp=True, increase_log_steps=True,
                 rescale=True, disabled=False, log_on_batch_idx=False, log_first_step=False,
                 log_images_kwargs=None):
        super().__init__()
        self.rescale = rescale
        self.batch_freq = batch_frequency
        self.max_images = max_images
        if not increase_log_steps:
            self.log_steps = [self.batch_freq]
        self.clamp = clamp
        self.disabled = disabled
        self.log_on_batch_idx = log_on_batch_idx
        self.log_images_kwargs = log_images_kwargs if log_images_kwargs else {}
        self.log_first_step = log_first_step

    @rank_zero_only
    def log_local(self, save_dir, split, images, global_step, current_epoch, batch_idx):
        root = os.path.join(save_dir, "image_log", split)
        for k in images:
            grid = torchvision.utils.make_grid(images[k], nrow=4)
            if self.rescale:
                grid = (grid + 1.0) / 2.0  # -1,1 -> 0,1; c,h,w
            grid = grid.transpose(0, 1).transpose(1, 2).squeeze(-1)
            grid = grid.numpy()
            grid = (grid * 255).astype(np.uint8)
            filename = "{}_gs-{:06}_e-{:06}_b-{:06}.png".format(k, global_step, current_epoch, batch_idx)
            path = os.path.join(root, filename)
            os.makedirs(os.path.split(path)[0], exist_ok=True)
            Image.fromarray(grid).save(path)

    def log_img(self, pl_module, batch, batch_idx, split="train"):
        check_idx = batch_idx  # if self.log_on_batch_idx else pl_module.global_step
        if (self.check_frequency(check_idx) and  # batch_idx % self.batch_freq == 0
                hasattr(pl_module, "log_images") and
                callable(pl_module.log_images) and
                self.max_images > 0):
            logger = type(pl_module.logger)

            is_train = pl_module.training
            if is_train:
                pl_module.eval()

            with torch.no_grad():
                images = pl_module.log_images(batch, split=split, **self.log_images_kwargs)

            for k in images:
                N = min(images[k].shape[0], self.max_images)
                images[k] = images[k][:N]
                if isinstance(images[k], torch.Tensor):
                    images[k] = images[k].detach().cpu()
                    if self.clamp:
                        images[k] = torch.clamp(images[k], -1., 1.)

            self.log_local(pl_module.logger.save_dir, split, images,
                           pl_module.global_step, pl_module.current_epoch, batch_idx)

            if is_train:
                pl_module.train()

    def check_frequency(self, check_idx):
        return check_idx % self.batch_freq == 0

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        if not self.disabled:
            self.log_img(pl_module, batch, batch_idx, split="train")



import copy
import os
from pathlib import Path

import cv2
import numpy as np
import torch
from torch import nn

from nataili.util.codeformer.facelib.detection.yolov5face.models.common import Conv
from nataili.util.codeformer.facelib.detection.yolov5face.models.yolo import Model
from nataili.util.codeformer.facelib.detection.yolov5face.utils.datasets import letterbox
from nataili.util.codeformer.facelib.detection.yolov5face.utils.general import (
    check_img_size,
    non_max_suppression_face,
    scale_coords,
    scale_coords_landmarks,
)

IS_HIGH_VERSION = tuple(map(int, torch.__version__.split("+")[0].split(".")[:3])) >= (
    1,
    9,
    0,
)


def isListempty(inList):
    if isinstance(inList, list):  # Is a list
        return all(map(isListempty, inList))
    return False  # Not a list


class YoloDetector:
    def __init__(
        self,
        config_name,
        min_face=10,
        target_size=None,
        device="cuda",
    ):
        """
        config_name: name of .yaml config with network configuration from models/ folder.
        min_face : minimal face size in pixels.
        target_size : target size of smaller image axis (choose lower for faster work). e.g. 480, 720, 1080.
                    None for original resolution.
        """
        self._class_path = Path(__file__).parent.absolute()
        self.target_size = target_size
        self.min_face = min_face
        self.detector = Model(cfg=config_name)
        self.device = device

    def _preprocess(self, imgs):
        """
        Preprocessing image before passing through the network. Resize and conversion to torch tensor.
        """
        pp_imgs = []
        for img in imgs:
            h0, w0 = img.shape[:2]  # orig hw
            if self.target_size:
                r = self.target_size / min(h0, w0)  # resize image to img_size
                if r < 1:
                    img = cv2.resize(img, (int(w0 * r), int(h0 * r)), interpolation=cv2.INTER_LINEAR)

            imgsz = check_img_size(max(img.shape[:2]), s=self.detector.stride.max())  # check img_size
            img = letterbox(img, new_shape=imgsz)[0]
            pp_imgs.append(img)
        pp_imgs = np.array(pp_imgs)
        pp_imgs = pp_imgs.transpose(0, 3, 1, 2)
        pp_imgs = torch.from_numpy(pp_imgs).to(self.device)
        pp_imgs = pp_imgs.float()  # uint8 to fp16/32
        return pp_imgs / 255.0  # 0 - 255 to 0.0 - 1.0

    def _postprocess(self, imgs, origimgs, pred, conf_thres, iou_thres):
        """
        Postprocessing of raw pytorch model output.
        Returns:
            bboxes: list of arrays with 4 coordinates of bounding boxes with format x1,y1,x2,y2.
            points: list of arrays with coordinates of 5 facial keypoints (eyes, nose, lips corners).
        """
        bboxes = [[] for _ in range(len(origimgs))]
        landmarks = [[] for _ in range(len(origimgs))]

        pred = non_max_suppression_face(pred, conf_thres, iou_thres)

        for image_id, origimg in enumerate(origimgs):
            img_shape = origimg.shape
            image_height, image_width = img_shape[:2]
            gn = torch.tensor(img_shape)[[1, 0, 1, 0]]  # normalization gain whwh
            gn_lks = torch.tensor(img_shape)[[1, 0, 1, 0, 1, 0, 1, 0, 1, 0]]  # normalization gain landmarks
            det = pred[image_id].cpu()
            scale_coords(imgs[image_id].shape[1:], det[:, :4], img_shape).round()
            scale_coords_landmarks(imgs[image_id].shape[1:], det[:, 5:15], img_shape).round()

            for j in range(det.size()[0]):
                box = (det[j, :4].view(1, 4) / gn).view(-1).tolist()
                box = list(
                    map(
                        int,
                        [
                            box[0] * image_width,
                            box[1] * image_height,
                            box[2] * image_width,
                            box[3] * image_height,
                        ],
                    )
                )
                if box[3] - box[1] < self.min_face:
                    continue
                lm = (det[j, 5:15].view(1, 10) / gn_lks).view(-1).tolist()
                lm = list(
                    map(
                        int,
                        [i * image_width if j % 2 == 0 else i * image_height for j, i in enumerate(lm)],
                    )
                )
                lm = [lm[i : i + 2] for i in range(0, len(lm), 2)]
                bboxes[image_id].append(box)
                landmarks[image_id].append(lm)
        return bboxes, landmarks

    def detect_faces(self, imgs, conf_thres=0.7, iou_thres=0.5):
        """
        Get bbox coordinates and keypoints of faces on original image.
        Params:
            imgs: image or list of images to detect faces on with BGR order (convert to RGB order for inference)
            conf_thres: confidence threshold for each prediction
            iou_thres: threshold for NMS (filter of intersecting bboxes)
        Returns:
            bboxes: list of arrays with 4 coordinates of bounding boxes with format x1,y1,x2,y2.
            points: list of arrays with coordinates of 5 facial keypoints (eyes, nose, lips corners).
        """
        # Pass input images through face detector
        images = imgs if isinstance(imgs, list) else [imgs]
        images = [cv2.cvtColor(img, cv2.COLOR_BGR2RGB) for img in images]
        origimgs = copy.deepcopy(images)

        images = self._preprocess(images)

        if IS_HIGH_VERSION:
            with torch.inference_mode():  # for pytorch>=1.9
                pred = self.detector(images)[0]
        else:
            with torch.no_grad():  # for pytorch<1.9
                pred = self.detector(images)[0]

        bboxes, points = self._postprocess(images, origimgs, pred, conf_thres, iou_thres)

        # return bboxes, points
        if not isListempty(points):
            bboxes = np.array(bboxes).reshape(-1, 4)
            points = np.array(points).reshape(-1, 10)
            padding = bboxes[:, 0].reshape(-1, 1)
            return np.concatenate((bboxes, padding, points), axis=1)
        else:
            return None

    def __call__(self, *args):
        return self.predict(*args)



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import time
from pathlib import Path

import torch
from diffusers import LMSDiscreteScheduler
from transformers import CLIPFeatureExtractor, CLIPTokenizer

from nataili.aitemplate import StableDiffusionAITPipeline
from nataili.cache import get_cache_directory
from nataili.model_manager.base import BaseModelManager
from nataili.util.logger import logger
from nataili.util.voodoo import init_ait_module


class AITemplateModelManager(BaseModelManager):
    def __init__(self, download_reference=True):
        super().__init__()
        self.download_reference = download_reference
        self.path = f"{get_cache_directory()}/aitemplate"
        self.models_db_name = "aitemplate"
        self.models_path = self.pkg / f"{self.models_db_name}.json"
        self.remote_db = (
            f"https://raw.githubusercontent.com/db0/AI-Horde-image-model-reference/main/{self.models_db_name}.json"
        )
        self.ait_workdir = None
        self.init()

    def init(self):
        if self.cuda_available:
            logger.info(f"Highest CUDA Compute Capability: {self.cuda_devices[0]['sm']}")
            logger.debug(f"Available CUDA Devices: {self.cuda_devices}")
            logger.info(f"Recommended GPU: {self.recommended_gpu}")
            sm = self.recommended_gpu[0]["sm"]
            logger.info(f"Using sm_{sm} for AITemplate")
            for aitemplate in self.models:
                ait_files = self.get_aitemplate_files(sm)
                if len(ait_files) > 0 and self.check_available(ait_files):
                    self.available_models.append(aitemplate)
                    logger.info(f"Available AITemplate: {aitemplate}")
            if len(self.available_models) == 0:
                logger.warning("No AITemplate available")
            else:
                self.ait_workdir = self.get_ait_workdir(sm)

    def download_ait(self, cuda_arch):
        """
        :param cuda_arch: CUDA Compute Capability
        Download AITemplate modules for CUDA Compute Capability
        """
        files = self.get_aitemplate_files(cuda_arch)
        download = self.get_aitemplate_download(cuda_arch)
        for i in range(len(download)):
            file_path = (
                f"{download[i]['file_path']}/{download[i]['file_name']}"
                if "file_path" in download[i]
                else files[i]["path"]
            )

            download_url = download[i]["file_url"]
            if not self.check_file_available(file_path):
                logger.debug(f"Downloading {download_url} to {file_path}")
                self.download_file(download_url, file_path)
        self.ait_workdir = self.get_ait_workdir(cuda_arch)

    def get_aitemplate_files(self, cuda_arch, model_name="stable_diffusion"):
        """
        :param cuda_arch: CUDA Compute Capability
        :param model_name: AITemplate model name
        :return: AITemplate files for CUDA Compute Capability
        """
        if cuda_arch == 89:
            return self.models[model_name]["config"]["sm89"]["files"]
        elif cuda_arch >= 80 and cuda_arch < 89:
            return self.models[model_name]["config"]["sm80"]["files"]
        elif cuda_arch == 75:
            return self.models[model_name]["config"]["sm75"]["files"]
        else:
            logger.warning("CUDA Compute Capability not supported")
            return []

    def get_aitemplate_download(self, cuda_arch, model_name="stable_diffusion"):
        """
        :param cuda_arch: CUDA Compute Capability
        :param model_name: AITemplate model name
        :return: AITemplate download details for CUDA Compute Capability
        """
        if cuda_arch == 89:
            return self.models[model_name]["config"]["sm89"]["download"]
        elif cuda_arch >= 80 and cuda_arch < 89:
            return self.models[model_name]["config"]["sm80"]["download"]
        elif cuda_arch == 75:
            return self.models[model_name]["config"]["sm75"]["download"]
        else:
            logger.warning("CUDA Compute Capability not supported")
            return []

    def get_ait_workdir(self, cuda_arch, model_name="stable_diffusion"):
        if cuda_arch == 89:
            return f"./{self.models[model_name]['config']['sm89']['download'][0]['file_path']}/"
        elif cuda_arch >= 80 and cuda_arch < 89:
            return f"./{self.models[model_name]['config']['sm80']['download'][0]['file_path']}/"
        elif cuda_arch == 75:
            return f"./{self.models[model_name]['config']['sm75']['download'][0]['file_path']}/"
        else:
            raise ValueError("CUDA Compute Capability not supported")

    def load(
        self,
        model_name: str = "stable_diffusion",
        gpu_id=0,
    ):
        """
        model_name: str. Name of the model to load. See available_models for a list of available models.
        gpu_id: int. The id of the cuda device to use.
        AITemplate requires CUDA.
        """
        if not self.cuda_available:
            logger.warning("AITemplate requires CUDA")
            return False
        if model_name not in self.models:
            logger.error(f"{model_name} not found")
            return False
        if len(self.available_models) == 0:
            logger.info("No available aitemplates")
            sm = self.recommended_gpu[0]["sm"]
            logger.init_ok(f"Downloading AITemplate for {sm}", status="Downloading")
            self.download_ait()
            logger.init_ok(f"AITemplate for {sm} downloaded", status="Downloading")
        if model_name not in self.loaded_models:
            tic = time.time()
            logger.init(f"{model_name}", status="Loading")
            self.loaded_models["ait"] = self.load_aitemplate(
                model_name,
                gpu_id=gpu_id,
            )
            logger.init_ok(f"Loading {model_name}", status="Success")
            toc = time.time()
            logger.init_ok(f"Loading {model_name}: Took {toc-tic} seconds", status="Success")
            return True

    def load_aitemplate(
        self,
        model_name,
        gpu_id=0,
    ):
        model_path = self.get_model_files(model_name)[0]["path"]
        model_path = f"{self.path}/{model_path}"
        device = torch.device(f"cuda:{gpu_id}")
        logger.info(f"Loading model {model_name} on {device}")
        logger.info(f"Model path: {model_path}")
        ait = {}
        ait["unet"] = init_ait_module("unet.so", self.ait_workdir)
        ait["clip"] = init_ait_module("clip.so", self.ait_workdir)
        ait["vae"] = init_ait_module("vae.so", self.ait_workdir)
        ait["pipe"] = StableDiffusionAITPipeline(
            vae=None,
            unet=None,
            text_encoder=None,
            tokenizer=CLIPTokenizer.from_pretrained("openai/clip-vit-large-patch14"),
            scheduler=LMSDiscreteScheduler.from_pretrained("runwayml/stable-diffusion-v1-5"),
            safety_checker=None,
            feature_extractor=CLIPFeatureExtractor.from_pretrained("openai/clip-vit-large-patch14"),
            clip_ait_exe=self.loaded_models["ait"]["clip"],
            unet_ait_exe=self.loaded_models["ait"]["unet"],
            vae_ait_exe=self.loaded_models["ait"]["vae"],
            filter_nsfw=False,
        ).to("cuda")
        return ait



from PIL import Image

from nataili.model_manager.sdu import SDUModelManager
from nataili.sdu import StableDiffusionUpscaler

mm = SDUModelManager()

mm.load("Stable Diffusion Upscaler")

sdu = StableDiffusionUpscaler(
    model=mm.loaded_models["Stable Diffusion Upscaler"]["model"],
    vae_840k=mm.loaded_models["Stable Diffusion Upscaler"]["vae_model_840k"],
    vae_560k=mm.loaded_models["Stable Diffusion Upscaler"]["vae_model_560k"],
    tokenizer=mm.loaded_models["Stable Diffusion Upscaler"]["tokenizer"],
    text_encoder=mm.loaded_models["Stable Diffusion Upscaler"]["text_encoder"],
    device=mm.loaded_models["Stable Diffusion Upscaler"]["device"]
)

input_image = Image.open("sd_2x_upscaler_demo.png").convert("RGB")

sdu(
    prompt = "the temple of fire by Ross Tran and Gerardo Dottori, oil on canvas",
    input_image = input_image,
)
    


import os

from PIL import Image
from tqdm import tqdm

from nataili.model_manager.clip import ClipModelManager
from nataili.util.logger import logger
from nataili.cache import Cache
from nataili.clip.image import ImageEmbed
from nataili.clip.text import TextEmbed


images = []

directory = "test_images"

mm = ClipModelManager()

mm.load("ViT-H-14")

cache = Cache(mm.loaded_models["ViT-H-14"]["cache_name"], cache_parentname="embeds", cache_subname="image")

image_embed = ImageEmbed(mm.loaded_models["ViT-H-14"], cache)

images = []

directory = "Y:/diffusiondb"

file_list = [os.path.splitext(file)[0] for file in os.listdir(directory) if file.endswith(".png") or file.endswith(".jpg") or file.endswith(".jpeg") or file.endswith(".webp")]

logger.info(f"Found {len(file_list)} images in directory")

filtered = cache.filter_list(file_list)

filtered = {key: None for key in filtered}

logger.info(f"Found {len(filtered)} images in cache")

batch_size = 256

for file in tqdm(os.listdir(directory)):
    if os.path.splitext(file)[0] in filtered:
        images.append({"filename": file})

logger.info(f"Found {len(images)} images to process")

for i in tqdm(range(0, len(images), batch_size)):
    batch = images[i:i+batch_size]
    for image in batch:
        image['pil_image'] = Image.open(f"{directory}/{image['filename']}").convert("RGB")
    image_embed._batch(batch)
    del batch
        






import uuid

from PIL import Image

from nataili.model_manager.compvis import CompVisModelManager
from nataili.stable_diffusion.compvis import CompVis
from nataili.util.logger import logger

init_image = Image.open(
    "./test_images/00726-2641044396_cute_girl_holding_a_giant_NVIDIA_gtx_1080ti_GPU_graphics_card,_Anime_Blu-Ray_boxart,_super_high_detail,_pixiv_trending.png"
).convert("RGB")

models = [
    {"name": "Future Diffusion", "trigger":"future style"},
    {"name": "Ultraskin", "trigger":"ultraskin"},
    {"name": "Concept Sheet", "trigger":"concept-art"},
    {"name": "CharHelper", "trigger":"CHV3CZombie"},
    {"name": "Pokemon3D", "trigger":"GMAX Mega <Pikachu>"},
    {"name": "Vector Art", "trigger":"vector-art"},
    {"name": "PRMJ", "trigger": ""},
]

prompt = "anime girl holding a giant nvidia gtx 1080ti gpu graphics card ### weird mouth,weird teeth,incorrect anatomy, bad anatomy, ugly, odd, hideous, unsightly, extra limbs, poorly drawn, poorly drawn face, bad drawing, sketch, disfigured, cropped, out of frame,fully clothed,sad,agony,pain,cut off,trimmed,headpiece, head necklace,out of frame, malformed,hideous face, too many limbs, missing fingers, too many fingers, text, bad drawing, sketch, incorrect anatomy, hideous, unsightly, extra limbs, poorly drawn, poorly drawn face, disfigured, cropped"

output_dir = f"./test_output/{str(uuid.uuid4())}"
mm = CompVisModelManager(download_reference=False)

for model in models:
    logger.info(f"Testing model: {model['name']}")
    mm.load(model["name"])
    compvis = CompVis(
        model=mm.loaded_models[model["name"]],
        model_name=model["name"],
        model_baseline=mm.models[model["name"]]["baseline"],
        output_dir=output_dir,
        disable_voodoo=True,
    )
    compvis.generate(
        f"{prompt} {model['trigger']}",
        init_img=init_image,
        height=768,
        width=768,
    )



# from .blip import blip_decoder
# from .cache import torch_gc
# from .cast import autocast_cpu, autocast_cuda
# from .check_prompt_length import check_prompt_length
# from .codeformer.codeformer import CodeFormer
# from .codeformer.face_restoration_helper import FaceRestoreHelper
# from .create_random_tensors import create_random_tensors
# from .get_next_sequence_number import get_next_sequence_number
# from .gfpgan.utils import GFPGANer
# from .image_grid import image_grid
# from .img2img import find_noise_for_image, get_matched_noise, process_init_mask, resize_image
# from .load_learned_embed_in_clip import load_learned_embed_in_clip
# from .load_list import load_list
# from .logger import logger, set_logger_verbosity
# from .normalized import normalized
# from .postprocessor import PostProcessor
# from .process_prompt_tokens import process_prompt_tokens
# from .save_sample import save_sample
# from .seed_to_int import seed_to_int
# from .switch import Switch



# based on https://github.com/isl-org/MiDaS

import cv2
import os
import torch
import torch.nn as nn
from torchvision.transforms import Compose

from .midas.dpt_depth import DPTDepthModel
from .midas.midas_net import MidasNet
from .midas.midas_net_custom import MidasNet_small
from .midas.transforms import Resize, NormalizeImage, PrepareForNet
from annotator.util import annotator_ckpts_path


ISL_PATHS = {
    "dpt_large": os.path.join(annotator_ckpts_path, "dpt_large-midas-2f21e586.pt"),
    "dpt_hybrid": os.path.join(annotator_ckpts_path, "dpt_hybrid-midas-501f0c75.pt"),
    "midas_v21": "",
    "midas_v21_small": "",
}

remote_model_path = "https://huggingface.co/lllyasviel/ControlNet/resolve/main/annotator/ckpts/dpt_hybrid-midas-501f0c75.pt"


def disabled_train(self, mode=True):
    """Overwrite model.train with this function to make sure train/eval mode
    does not change anymore."""
    return self


def load_midas_transform(model_type):
    # https://github.com/isl-org/MiDaS/blob/master/run.py
    # load transform only
    if model_type == "dpt_large":  # DPT-Large
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "dpt_hybrid":  # DPT-Hybrid
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "midas_v21":
        net_w, net_h = 384, 384
        resize_mode = "upper_bound"
        normalization = NormalizeImage(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    elif model_type == "midas_v21_small":
        net_w, net_h = 256, 256
        resize_mode = "upper_bound"
        normalization = NormalizeImage(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    else:
        assert False, f"model_type '{model_type}' not implemented, use: --model_type large"

    transform = Compose(
        [
            Resize(
                net_w,
                net_h,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method=resize_mode,
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            normalization,
            PrepareForNet(),
        ]
    )

    return transform


def load_model(model_type):
    # https://github.com/isl-org/MiDaS/blob/master/run.py
    # load network
    model_path = ISL_PATHS[model_type]
    if model_type == "dpt_large":  # DPT-Large
        model = DPTDepthModel(
            path=model_path,
            backbone="vitl16_384",
            non_negative=True,
        )
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "dpt_hybrid":  # DPT-Hybrid
        if not os.path.exists(model_path):
            from basicsr.utils.download_util import load_file_from_url
            load_file_from_url(remote_model_path, model_dir=annotator_ckpts_path)

        model = DPTDepthModel(
            path=model_path,
            backbone="vitb_rn50_384",
            non_negative=True,
        )
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "midas_v21":
        model = MidasNet(model_path, non_negative=True)
        net_w, net_h = 384, 384
        resize_mode = "upper_bound"
        normalization = NormalizeImage(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )

    elif model_type == "midas_v21_small":
        model = MidasNet_small(model_path, features=64, backbone="efficientnet_lite3", exportable=True,
                               non_negative=True, blocks={'expand': True})
        net_w, net_h = 256, 256
        resize_mode = "upper_bound"
        normalization = NormalizeImage(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )

    else:
        print(f"model_type '{model_type}' not implemented, use: --model_type large")
        assert False

    transform = Compose(
        [
            Resize(
                net_w,
                net_h,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method=resize_mode,
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            normalization,
            PrepareForNet(),
        ]
    )

    return model.eval(), transform


class MiDaSInference(nn.Module):
    MODEL_TYPES_TORCH_HUB = [
        "DPT_Large",
        "DPT_Hybrid",
        "MiDaS_small"
    ]
    MODEL_TYPES_ISL = [
        "dpt_large",
        "dpt_hybrid",
        "midas_v21",
        "midas_v21_small",
    ]

    def __init__(self, model_type):
        super().__init__()
        assert (model_type in self.MODEL_TYPES_ISL)
        model, _ = load_model(model_type)
        self.model = model
        self.model.train = disabled_train

    def forward(self, x):
        with torch.no_grad():
            prediction = self.model(x)
        return prediction




"""
Small modification of CodeFormer from  repo.
Uses RealESRGAN x2 from ESRGAN ModelManager.
Allows cache dir to be specified.
Allows device to be specified. NOTE: Device selection seems funky, high CPU usage even when using cuda.
Uses modified FaceRestoreHelper.
"""
import cv2
import numpy as np
import torch
from basicsr.archs.rrdbnet_arch import RRDBNet
from basicsr.utils import img2tensor, tensor2img
from basicsr.utils.registry import ARCH_REGISTRY
from PIL import Image
from realesrgan import RealESRGANer
from torchvision.transforms.functional import normalize

from nataili.util.codeformer.codeformer_arch import CodeFormer as CodeFormer_Arch
from nataili.util.codeformer.face_restoration_helper import FaceRestoreHelper
from nataili.util.codeformer.misc import is_gray


class CodeFormer(torch.nn.Module):
    def __init__(
        self,
        esrgan_model_manager,
        gfpgan_model_manager,
        codeformer_model_manager,
        upscale=2,
        detection_model="retinaface_resnet50",
        bg_upsampler=None,
        bg_tile=400,
        device="cuda",
    ):
        """
        Args:
            weights (str): path to the pretrained model
            upscale (int): upscale factor
            detection_model (str): Choices: retinaface_resnet50, retinaface_mobile0.25, YOLOv5l, YOLOv5n. Default: retinaface_resnet50
            bg_upsampler (str): Choices: RealESRGAN, None. Default: None
            bg_tile (int): tile size for background upsampling. Default: 400
        """
        super().__init__()
        self.esrgan_model_manager = esrgan_model_manager
        self.gfpgan_model_manager = gfpgan_model_manager
        self.upscale = upscale
        self.detection_model = detection_model
        self.bg_tile = bg_tile

        if bg_upsampler == "realesrgan":
            model = RRDBNet(
                num_in_ch=3,
                num_out_ch=3,
                num_feat=64,
                num_block=23,
                num_grow_ch=32,
                scale=2,
            )
            model_path = esrgan_model_manager.get_model_files("RealESRGAN_x2plus")[0]["path"]
            if "RealESRGAN_x2plus" not in esrgan_model_manager.available_models:
                esrgan_model_manager.download_model("RealESRGAN_x2plus")
            upsampler = RealESRGANer(
                scale=2,
                model_path=model_path,
                model=model,
                tile=bg_tile,
                tile_pad=40,
                pre_pad=0,
                half=True if torch.cuda.is_available() else False,
                device=device,
            )
            self.bg_upsampler = upsampler
        else:
            self.bg_upsampler = None

        self.model = CodeFormer_Arch(
            dim_embd=512,
            codebook_size=1024,
            n_head=8,
            n_layers=9,
            connect_list=["32", "64", "128", "256"],
        )

        model_path = codeformer_model_manager.get_model_files("CodeFormers")[0]["path"]
        model_path = f"{codeformer_model_manager.path}/{model_path}"

        checkpoint = torch.load(model_path, map_location="cpu")["params_ema"]
        self.model.load_state_dict(checkpoint)
        self.model.eval().requires_grad_(False)

        self.face_helper = FaceRestoreHelper(
            self.upscale,
            face_size=512,
            crop_ratio=(1, 1),
            det_model=self.detection_model,
            save_ext="png",
            use_parse=True,
            device=device,
            model_rootpath=gfpgan_model_manager.path,  # GFPGAN uses the same FaceRestoreHelper models
            source_facefixer="CodeFormer",
        )

    @property
    def device(self):
        return next(iter(self.parameters())).device

    def to(self, device):
        super().to(device)
        self.face_helper.face_det.to(device)
        self.face_helper.face_parse.to(device)
        self.face_helper.device = device
        return self

    def forward(
        self,
        pil_image: Image.Image,
        fidelity_weight=0.5,
        has_aligned=False,
        only_center_face=False,
        draw_face_bounding_box=False,
    ) -> Image.Image:
        self.face_helper.all_landmarks_5 = []
        self.face_helper.det_faces = []
        self.face_helper.affine_matrices = []
        self.face_helper.inverse_affine_matrices = []
        self.face_helper.cropped_faces = []
        self.face_helper.restored_faces = []
        self.face_helper.pad_input_imgs = []
        img = np.array(pil_image)

        if has_aligned:
            # the input faces are already cropped and aligned
            img = cv2.resize(img, (512, 512), interpolation=cv2.INTER_LINEAR)
            self.face_helper.is_gray = is_gray(img, threshold=10)
            self.face_helper.cropped_faces = [img]
        else:
            self.face_helper.read_image(img)
            # get face landmarks for each face
            self.face_helper.get_face_landmarks_5(only_center_face=only_center_face, resize=640, eye_dist_threshold=5)
            self.face_helper.align_warp_face()

        # face restoration for each cropped face
        for idx, cropped_face in enumerate(self.face_helper.cropped_faces):
            # prepare data
            cropped_face_t = img2tensor(cropped_face / 255.0, bgr2rgb=True, float32=True)
            normalize(cropped_face_t, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), inplace=True)
            cropped_face_t = cropped_face_t.unsqueeze(0).to(self.device)

            try:
                with torch.no_grad():
                    output = self.model(cropped_face_t, w=fidelity_weight, adain=True)[0]
                    restored_face = tensor2img(output, rgb2bgr=True, min_max=(-1, 1))
                del output
                torch.cuda.empty_cache()
            except Exception as error:
                print(f"\tFailed inference for CodeFormer: {error}")
                restored_face = tensor2img(cropped_face_t, rgb2bgr=True, min_max=(-1, 1))

            restored_face = restored_face.astype("uint8")
            self.face_helper.add_restored_face(restored_face)

        # paste_back
        if not has_aligned:
            # upsample the background
            if self.bg_upsampler is not None:
                # Now only support RealESRGAN for upsampling background
                bg_img = self.bg_upsampler.enhance(img, outscale=self.upscale)[0]
            else:
                bg_img = None
            self.face_helper.get_inverse_affine(None)
            # paste each restored face to the input image
            restored_img = self.face_helper.paste_faces_to_input_image(
                upsample_img=bg_img, draw_box=draw_face_bounding_box
            )

        return Image.fromarray(restored_img)


def test_codeformer():
    model = CodeFormer().to(torch.device("cuda" if torch.cuda.is_available() else "cpu"))
    input_image = Image.open("tests/test.png")
    model(input_image).save("tests/output.png")


def test_codeformer_cpu():
    model = CodeFormer().to(torch.device("cpu"))
    input_image = Image.open("tests/test.png")
    model(input_image).save("tests/output.png")



from nataili.model_manager.compvis import CompVisModelManager
from nataili.util.logger import logger

mm = CompVisModelManager()

for model in mm.models:
    logger.info(f"Downloading {model}...")
    mm.download_model(model)



from PIL import Image

from nataili.blip import Caption
from nataili.model_manager.blip import BlipModelManager
from nataili.util.logger import logger

image = Image.open("01.png")

mm = BlipModelManager()

mm.load("BLIP")

blip = Caption(mm.loaded_models["BLIP"])

logger.generation(f"caption: {blip(image)}")



import os

from PIL import Image

from nataili.model_manager.clip import ClipModelManager
from nataili.clip.interrogate import Interrogator
from nataili.util.logger import logger


images = []

directory = "test_images"

mm = ClipModelManager()

mm.load("ViT-L/14")

interrogator = Interrogator(
    mm.loaded_models["ViT-L/14"],
)

for file in os.listdir(directory):
    image  = Image.open(f"{directory}/{file}").convert("RGB")
    results = interrogator(image=image, text_array=None, rank=True, top_count=5)
    """
    or
    results = interrogator(filename=file, directory=directory, text_array=None, rank=True, top_count=5)
    """
    logger.generation(results)



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import json
import os
import shutil
import sys
import zipfile
from pathlib import Path
from uuid import uuid4

import git
import requests
import torch
from tqdm import tqdm
from transformers import logging

from nataili import disable_download_progress
from nataili.cache import get_cache_directory
from nataili.util.logger import logger

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources

logging.set_verbosity_error()


class BaseModelManager:
    def __init__(self, download_reference=True):
        self.path = get_cache_directory()
        self.models = {}
        self.available_models = []
        self.loaded_models = {}
        self.tainted_models = []
        self.pkg = importlib_resources.files("nataili")
        self.models_db_name = "models"
        self.models_path = self.pkg / f"{self.models_db_name}.json"
        self.cuda_available = torch.cuda.is_available()
        self.cuda_devices, self.recommended_gpu = self.get_cuda_devices()
        self.remote_db = (
            f"https://raw.githubusercontent.com/db0/AI-Horde-image-model-reference/main/{self.models_db_name}.json"
        )
        self.download_reference = download_reference

    def init(self, list_models=False):
        if self.download_reference:
            self.models = self.download_model_reference()
            logger.info(f"Downloaded model reference. Got {len(self.models)} models.")
        else:
            self.models = json.loads((self.models_path).read_text())
            logger.info(f"Loaded model reference. Got {len(self.models)} models.")
        if list_models:
            for model in self.models:
                logger.info(model)
        models_available = []
        for model in self.models:
            if self.check_model_available(model):
                models_available.append(model)
        self.available_models = models_available
        logger.info(f"Got {len(self.available_models)} available models.")
        if list_models:
            for model in self.available_models:
                logger.info(model)

    def download_model_reference(self):
        try:
            logger.init("Model Reference", status="Downloading")
            response = requests.get(self.remote_db)
            logger.init_ok("Model Reference", status="OK")
            models = response.json()
            return models
        except Exception as e:
            logger.init_err("Model Reference", status=f"Download failed: {e}")
            logger.init_warn("Model Reference", status="Local")
            return json.loads((self.models_path).read_text())

    def get_model(self, model_name):
        return self.models.get(model_name)

    def get_model_files(self, model_name):
        """
        :param model_name: Name of the model
        Returns the files for a model
        """
        if self.models[model_name]["type"] == "diffusers":
            return []
        return self.models[model_name]["config"]["files"]

    def get_model_download(self, model_name):
        """
        :param model_name: Name of the model
        Returns the download details for a model
        """
        return self.models[model_name]["config"]["download"]

    def get_available_models(self):
        """
        Returns the available models
        """
        return self.available_models

    def get_available_models_by_types(self, model_types=None):
        if not model_types:
            model_types = ["ckpt", "diffusers"]
        models_available = []
        for model in self.models:
            if self.models[model]["type"] in model_types and self.check_available(self.get_model_files(model)):
                models_available.append(model)
        return models_available

    def count_available_models_by_types(self, model_types=None):
        return len(self.get_available_models_by_types(model_types))

    def get_loaded_models(self):
        """
        Returns the loaded models
        """
        return self.loaded_models

    def get_loaded_model(self, model_name):
        """
        :param model_name: Name of the model
        Returns the loaded model
        """
        return self.loaded_models[model_name]

    def get_loaded_models_names(self, string=False):
        """
        :param string: If True, returns concatenated string of model names
        Returns a list of the loaded model names
        """
        if string:
            return ", ".join(self.loaded_models.keys())
        return list(self.loaded_models.keys())

    def is_model_loaded(self, model_name):
        """
        :param model_name: Name of the model
        Returns whether the model is loaded
        """
        return model_name in self.loaded_models

    def unload_model(self, model_name):
        """
        :param model_name: Name of the model
        Unloads a model
        """
        if model_name in self.loaded_models:
            del self.loaded_models[model_name]
            return True
        return False

    def unload_all_models(self):
        """
        Unloads all models
        """
        for model in self.loaded_models:
            del self.loaded_models[model]
        return True

    def taint_model(self, model_name):
        """Marks a model as not valid by remiving it from available_models"""
        if model_name in self.available_models:
            self.available_models.remove(model_name)
            self.tainted_models.append(model_name)

    def taint_models(self, models):
        for model in models:
            self.taint_model(model)

    def validate_model(self, model_name, skip_checksum=False):
        """
        :param model_name: Name of the model
        :param skip_checksum: If True, skips checksum validation
        For each file in the model, checks if the file exists and if the checksum is correct
        Returns True if all files are valid, False otherwise
        """
        files = self.get_model_files(model_name)
        logger.debug(f"Validating {model_name} with {len(files)} files")
        logger.debug(files)
        for file_details in files:
            if ".yaml" in file_details["path"]:
                continue
            if not self.check_file_available(file_details["path"]):
                logger.debug(f"File {file_details['path']} not found")
                return False
            if not skip_checksum and not self.validate_file(file_details):
                logger.debug(f"File {file_details['path']} has invalid checksum")
                return False
        return True

    @staticmethod
    def get_file_md5sum_hash(file_name):
        # Bail out if the source file doesn't exist
        if not os.path.isfile(file_name):
            return

        # Check if we have a cached md5 hash for the source file
        # and use that unless our source file is newer than our hash
        md5_file = f"{os.path.splitext(file_name)[0]}.md5"
        source_timestamp = os.path.getmtime(file_name)
        if os.path.isfile(md5_file):
            hash_timestamp = os.path.getmtime(md5_file)
        else:
            hash_timestamp = 0
        if hash_timestamp > source_timestamp:
            # Use our cached hash
            with open(md5_file, "rt") as handle:
                md5_hash = handle.read().split()[0]
            return md5_hash

        # Calculate the hash of the source file
        with open(file_name, "rb") as file_to_check:
            file_hash = hashlib.md5()
            while True:
                chunk = file_to_check.read(8192)  # Changed just because it broke pylint
                if not chunk:
                    break
                file_hash.update(chunk)
        md5_hash = file_hash.hexdigest()

        # Cache this md5 hash we just calculated. Use md5sum format files
        # so we can also use OS tools to manipulate these md5 files
        try:
            with open(md5_file, "wt") as handle:
                handle.write(f"{md5_hash} *{os.path.basename(md5_file)}")
        except (OSError, PermissionError):
            logger.debug("Could not write to md5sum file, ignoring")

        return md5_hash

    @staticmethod
    def get_file_sha256_hash(file_name):
        if not os.path.isfile(file_name):
            raise FileNotFoundError("No file {}".format(file_name))

        # Check if we have a cached sha256 hash for the source file
        # and use that unless our source file is newer than our hash
        sha256_file = f"{os.path.splitext(file_name)[0]}.sha256"
        source_timestamp = os.path.getmtime(file_name)
        if os.path.isfile(sha256_file):
            hash_timestamp = os.path.getmtime(sha256_file)
        else:
            hash_timestamp = 0
        if hash_timestamp > source_timestamp:
            # Use our cached hash
            with open(sha256_file, "rt") as handle:
                sha256_hash = handle.read().split()[0]
            return sha256_hash

        # Calculate the hash of the source file
        with open(file_name, "rb") as file_to_check:
            file_hash = hashlib.sha256()
            while True:
                chunk = file_to_check.read(8192)
                if not chunk:
                    break
                file_hash.update(chunk)
        sha256_hash = file_hash.hexdigest()

        # Cache this sha256 hash we just calculated. Use sha256sum format files
        # so we can also use OS tools to manipulate these md5 files
        try:
            with open(sha256_file, "wt") as handle:
                handle.write(f"{sha256_hash} *{os.path.basename(sha256_file)}")
        except (OSError, PermissionError):
            logger.debug("Could not write to sha256sum file, ignoring")

        return sha256_hash

    def validate_file(self, file_details):
        """
        :param file_details: A single file from the model's files list
        Checks if the file exists and if the checksum is correct
        Returns True if the file is valid, False otherwise
        """
        full_path = f"{self.path}/{file_details['path']}"

        # Default to sha256 hashes
        if "sha256sum" in file_details:
            logger.debug(f"Getting sha256sum of {full_path}")
            sha256_file_hash = self.get_file_sha256_hash(full_path)
            logger.debug(f"sha256sum: {sha256_file_hash}")
            logger.debug(f"Expected: {file_details['sha256sum']}")
            if file_details["sha256sum"] != sha256_file_hash:
                return False
            else:
                return True

        # If sha256 is not available, fall back to md5
        if "md5sum" in file_details:
            logger.debug(f"Getting md5sum of {full_path}")
            md5_file_hash = self.get_file_md5sum_hash(full_path)
            logger.debug(f"md5sum: {md5_file_hash}")
            logger.debug(f"Expected: {file_details['md5sum']}")
            if file_details["md5sum"] != md5_file_hash:
                return False
            else:
                return True

        # If no hashes available, return True for now
        # THIS IS A SECURITY RISK, EVENTUALLY WE SHOULD RETURN FALSE
        # But currently not all models specify hashes
        return True

    def check_file_available(self, file_path):
        """
        :param file_path: Path of the model's file. File is from the model's files list.
        Checks if the file exists
        Returns True if the file exists, False otherwise
        """
        full_path = f"{self.path}/{file_path}"
        return os.path.exists(full_path)

    def check_available(self, files):
        """
        :param files: List of files from the model's files list
        Checks if all files exist
        Returns True if all files exist, False otherwise
        """
        available = True
        for file in files:
            if ".yaml" in file["path"]:
                continue
            if not self.check_file_available(file["path"]):
                available = False
        return available

    def download_file(self, url, file_path):
        """
        :param url: URL of the file to download. URL is from the model's download list.
        :param file_path: Path of the model's file. File is from the model's files list.
        Downloads a file
        """
        full_path = f"{self.path}/{file_path}"
        os.makedirs(os.path.dirname(full_path), exist_ok=True)
        pbar_desc = full_path.split("/")[-1]
        r = requests.get(url, stream=True, allow_redirects=True)
        with open(full_path, "wb") as f:
            with tqdm(
                # all optional kwargs
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                miniters=1,
                desc=pbar_desc,
                total=int(r.headers.get("content-length", 0)),
                disable=disable_download_progress.active,
            ) as pbar:
                for chunk in r.iter_content(chunk_size=16 * 1024):
                    if chunk:
                        f.write(chunk)
                        pbar.update(len(chunk))

    def download_model(self, model_name):
        """
        :param model_name: Name of the model
        Checks if the model is available, downloads the model if it is not available.
        After download, validates the model.
        Returns True if the model is available, False otherwise.

        Supported download types:
        - http(s) (url)
        - git (repo url)

        Other:
        - write content to file
        - symlink file
        - delete file
        - unzip file
        """
        if model_name in self.available_models and model_name not in self.tainted_models:
            logger.info(f"{model_name} is already available.")
            return True
        download = self.get_model_download(model_name)
        files = self.get_model_files(model_name)
        for i in range(len(download)):
            file_path = (
                f"{download[i]['file_path']}/{download[i]['file_name']}"
                if "file_path" in download[i]
                else files[i]["path"]
            )

            if "file_url" in download[i]:
                download_url = download[i]["file_url"]
            if "file_name" in download[i]:
                download_name = download[i]["file_name"]
            if "file_path" in download[i]:
                download_path = download[i]["file_path"]

            if "manual" in download[i]:
                logger.warning(
                    f"The model {model_name} requires manual download from {download_url}. "
                    f"Please place it in {download_path}/{download_name} then press ENTER to continue..."
                )
                input("")
                continue
            # TODO: simplify
            if "file_content" in download[i]:
                file_content = download[i]["file_content"]
                logger.info(f"writing {file_content} to {file_path}")
                os.makedirs(os.path.join(self.path, download_path), exist_ok=True)
                with open(os.path.join(self.path, os.path.join(download_path, download_name)), "w") as f:
                    f.write(file_content)
            elif "symlink" in download[i]:
                logger.info(f"symlink {file_path} to {download[i]['symlink']}")
                symlink = download[i]["symlink"]
                os.makedirs(os.path.join(self.path, download_path), exist_ok=True)
                os.symlink(symlink, os.path.join(self.path, os.path.join(download_path, download_name)))
            elif "git" in download[i]:
                logger.info(f"git clone {download_url} to {file_path}")
                os.makedirs(os.path.join(self.path, file_path), exist_ok=True)
                git.Git(os.path.join(self.path, file_path)).clone(download_url)
            elif "unzip" in download[i]:
                zip_path = f"{self.path}/{download_name}.zip"
                temp_path = f"{self.path}/{str(uuid4())}/"
                os.makedirs(temp_path, exist_ok=True)
                self.download_file(download_url, zip_path)
                logger.info(f"unzip {zip_path}")
                with zipfile.ZipFile(zip_path, "r") as zip_ref:
                    zip_ref.extractall(temp_path)
                logger.info(f"moving {temp_path} to {download_path}")
                shutil.move(temp_path, os.path.join(self.path, download_path))
                logger.info(f"delete {zip_path}")
                os.remove(zip_path)
                logger.info(f"delete {temp_path}")
                shutil.rmtree(temp_path)
            else:
                if not self.check_file_available(file_path) or model_name in self.tainted_models:
                    logger.debug(f"Downloading {download_url} to {file_path}")
                    self.download_file(download_url, file_path)
        if not self.validate_model(model_name):
            return False
        self.init()
        return True

    def download_all_models(self):
        """
        Downloads all models
        """
        for model in self.get_filtered_model_names(download_all=True):
            if not self.check_model_available(model):
                logger.init(f"{model}", status="Downloading")
                self.download_model(model)
            else:
                logger.info(f"{model} is already downloaded.")
        return True

    def check_model_available(self, model_name):
        """
        :param model_name: Name of the model
        Checks if the model is available.
        Returns True if the model is available, False otherwise.
        """
        if model_name not in self.models:
            return False
        return self.check_available(self.get_model_files(model_name))

    def get_cuda_devices(self):
        """
        Checks if CUDA is available.
        If CUDA is available, it returns a list of all available CUDA devices.
        If CUDA is not available, it returns an empty list.
        CUDA Device info: id, name, sm
        List is sorted by sm (compute capability) in descending order.
        Also returns the recommended GPU (highest sm).
        """

        if torch.cuda.is_available():
            number_of_cuda_devices = torch.cuda.device_count()
            cuda_arch = []
            for i in range(number_of_cuda_devices):
                cuda_device = {
                    "id": i,
                    "name": torch.cuda.get_device_name(i),
                    "sm": torch.cuda.get_device_capability(i)[0] * 10 + torch.cuda.get_device_capability(i)[1],
                }
                cuda_arch.append(cuda_device)
            cuda_arch = sorted(cuda_arch, key=lambda k: k["sm"], reverse=True)
            recommended_gpu = [x for x in cuda_arch if x["sm"] == cuda_arch[0]["sm"]]
            return cuda_arch, recommended_gpu
        else:
            return None, None

    def get_filtered_models(self, **kwargs):
        """
        Get all models
        :param kwargs: filter based on metadata of the model reference db
        :return: list of models
        """
        filtered_models = self.models
        for keyword in kwargs:
            iterating_models = filtered_models.copy()
            filtered_models = {}
            for model in iterating_models:
                if iterating_models[model].get(keyword) == kwargs[keyword]:
                    filtered_models[model] = iterating_models[model]
        return filtered_models

    def get_filtered_model_names(self, **kwargs):
        """
        Get all model names
        :param kwargs: filter based on metadata of the model reference db
        :return: list of model names
        """
        filtered_models = self.get_filtered_models(**kwargs)
        return list(filtered_models.keys())



from nataili.aitemplate import AITemplate
from nataili.model_manager.aitemplate import AITemplateModelManager
from nataili.util.logger import logger

mm = AITemplateModelManager()

mm.load()


def run():
    while True:
        logger.info("init")
        ait = AITemplate(mm.loaded_models["ait"])
        logger.info("start")
        ait.generate("corgi", ddim_steps=30)
        logger.info("end")


if __name__ == "__main__":
    run()
